

import warnings

import numpy as np
import scipy.interpolate as scpinterp
import scipy.stats as scpstats
import matplotlib.pyplot as plt


_LTYPES = [int, float, np.int_, np.float64]
_MISCUT = True


# ###############################################
#           utility
# ###############################################


def _check_bool(var, vardef=None, varname=None):
    if var is None:
        var = vardef
    if not isinstance(var, bool):
        msg = (
            "Arg {} must be a bool\n".format(varname)
            + "  You provided: {}".format(type(var))
        )
        raise Exception(msg)
    return var


def _are_broadcastable(**kwdargs):

    # Check if broadcastable
    lv = list(kwdargs.values())
    c0 = (
        all([isinstance(vv, np.ndarray) for vv in lv])
        and all([
            all([
                (m == n) or (m == 1) or (n == 1)
                for m, n in zip(vv.shape[::-1], lv[0].shape[::-1])
            ])
            for vv in lv
        ])
    )

    # raise Exception if strict
    if not c0:
        msg = (
            "All args must be broadcastable with each other!\n"
            + "You provided:\n"
            + "\n".join([
                '\t- {}.shape = {}'.format(k0, v0.shape)
                for k0, v0 in kwdargs.items()
            ])
        )
        raise Exception(msg)


# ###############################################
# ###############################################
#           CrystalBragg
# ###############################################
# ###############################################


def _checkformat_xixj(xi, xj):

    if xi is None or xj is None:
        msg = (
            "Arg xi and xj must be provided!\n"
            "Provided:\n"
            f"\t- xi: {xi}\n"
            f"\t- xj: {xj}\n"
        )
        raise Exception(msg)

    xi = np.atleast_1d(xi)
    xj = np.atleast_1d(xj)

    if xi.shape == xj.shape:
        return xi, xj, (xi, xj)
    else:
        return xi, xj, np.meshgrid(
            xi, xj,
            copy=True, sparse=False, indexing='ij',
        )


# ###############################################
#           sampling
# ###############################################


def _check_dthetapsi(
    dtheta=None, psi=None,
    extenthalf_psi=None, extenthalf_dtheta=None,
    ntheta=None, npsi=None,
    include_summit=None,
):
    """ Return formatted dtheta and psi

    They are returned with the same shape (at least 1d arrays)

    They can be:
        - 'envelop': if psi of dtheta = 'envelop', psi and dtheta are computed
            to describe the contour of the crystal as 2 (nenvelop,) arrays
            The nvelop is computed from (npsi, ntheta)
        - np.ndarrays or scalar: the routine just converts to np.ndarrays
            (using np.atleast_1d()) and checks they have the same shape

    """

    # Check inputs
    if dtheta is None:
        dtheta = 0.
    if psi is None:
        psi = 0.

    # if envelop => get all points around cryst + summit
    if any([isinstance(vv, str) and vv == 'envelop' for vv in [dtheta, psi]]):
        psi, dtheta = CrystBragg_sample_outline_sphrect(
            extenthalf_psi, extenthalf_dtheta,
            npsi=npsi, ntheta=ntheta,
            include_summit=include_summit,
        )

    c0 = all([
        type(vv) in _LTYPES
        or isinstance(vv, np.ndarray)
        for vv in [dtheta, psi]
    ])

    dtheta = np.atleast_1d(dtheta)
    psi = np.atleast_1d(psi)
    if psi.shape != dtheta.shape:
        msg = (
            "dtheta and psi should have the same shape\n"
            + "\t- dtheta.shape = {}\n".format(dtheta.shape)
            + "\t- psi.shape = {}".format(psi.shape)
        )
        raise Exception(msg)
    return dtheta, psi


def CrystBragg_sample_outline_sphrect(
    extent_psi, extent_dtheta,
    npsi=None, ntheta=None,
    include_summit=None,
):
    """ Return psi, dtheta describing the envelop of a crystal

    They are computed from
        - extent_psi, extent_dtheta: np.ndarrays for size 2
        - npsi, ntheta: integers

    They are returned with the same shape:
        (nenvelop,) arrays describing the contour of the crystal

    Optionally, the crystal summit can be appended at the end

    """

    # check inputs
    if include_summit is None:
        include_summit = True
    if ntheta is None:
        ntheta = 5
    if npsi is None:
        npsi = 3

    # compute
    psi = extent_psi*np.linspace(-1, 1., npsi)
    dtheta = extent_dtheta*np.linspace(-1, 1., ntheta)
    psimin = np.full((ntheta,), psi[0])
    psimax = np.full((ntheta,), psi[-1])
    dthetamin = np.full((npsi,), dtheta[0])
    dthetamax = np.full((npsi,), dtheta[-1])
    psi = np.concatenate((psi, psimax, psi[::-1], psimin))
    dtheta = np.concatenate((dthetamin, dtheta, dthetamax, dtheta[::-1]))
    if include_summit is True:
        psi = np.r_[psi, 0.]
        dtheta = np.r_[dtheta, 0.]
    return psi, dtheta


def CrystBragg_get_noute1e2_from_psitheta(
    nout, e1, e2,
    psi=None, dtheta=None,
    e1e2=None, sameshape=None,
    extenthalf_psi=None,
    extenthalf_dtheta=None,
    ntheta=None, npsi=None,
    include_summit=None,
):
    """ Return local unit vectors at chosen points on the crystal surface

    The points are defined by (psi, dtheta), which have to be the same shape
    and can be:
        - arbitrary: must be the same shape, can have up to 4 dimensions
        - defined from the envelop using (nspi, ntheta) fed to
            CrystBragg_sample_outline_sphrect()
            In this case psi and theta are 1d

    Local unit vectors (vout, ve1, ve2) at each point are defined the global
    unit vectors (taken at the crystal summit) passed as input (nout, e1, e2)

    Return
    ------
    vout:   np.ndarray
        (X, Y, Z) coordinates of nout normal outwards local vectors
    ve1:    np.ndarray
        (X, Y, Z) coordinates of e1 local tangential vectors
        Only returned if e1e2 is True
    ve2:    np.ndarray
        (X, Y, Z) coordinates of e2 local tangential vectors
        Only returned if e1e2 is True

    In all cases, the shape of the unit vectors is (3, psi.shape)

    """

    # check inputs
    if e1e2 is None:
        e1e2 = True
    if sameshape is None:
        sameshape = psi.shape == nout.shape[1:]
    if sameshape:
        assert psi.shape == nout.shape[1:]

    dtheta, psi = _check_dthetapsi(
        dtheta=dtheta, psi=psi,
        extenthalf_psi=extenthalf_psi,
        extenthalf_dtheta=extenthalf_dtheta,
        ntheta=ntheta, npsi=npsi,
        include_summit=include_summit,
    )

    # Prepare
    if sameshape is False:
        assert psi.ndim in [1, 2, 3, 4]
        if psi.ndim == 1:
            nout = nout[:, None]
            e1, e2 = e1[:, None], e2[:, None]
        elif psi.ndim == 2:
            nout = nout[:, None, None]
            e1, e2 = e1[:, None, None], e2[:, None, None]
        elif psi.ndim == 3:
            nout = nout[:, None, None, None]
            e1, e2 = e1[:, None, None, None], e2[:, None, None, None]
        else:
            nout = nout[:, None, None, None, None]
            e1 = e1[:, None, None, None, None]
            e2 = e2[:, None, None, None, None]

    # Not necessary for broadcasting (last dims first)
    theta = dtheta  # + np.pi/2.
    # psi = psi[None, ...]

    # Compute
    # vout = (
    # (np.cos(psi)*nout + np.sin(psi)*e1)*np.sin(theta) + np.cos(theta)*e2
    # )
    vout = (
         (np.cos(psi)*nout + np.sin(psi)*e1)*np.cos(theta) + np.sin(theta)*e2
         )
    if e1e2:
        ve1 = -np.sin(psi)*nout + np.cos(psi)*e1
        ve2 = np.array([vout[1, ...]*ve1[2, ...] - vout[2, ...]*ve1[1, ...],
                        vout[2, ...]*ve1[0, ...] - vout[0, ...]*ve1[2, ...],
                        vout[0, ...]*ve1[1, ...] - vout[1, ...]*ve1[0, ...]])
        return vout, ve1, ve2
    else:
        return vout


def CrystBragg_sample_outline_plot_sphrect(
    center, nout, e1, e2,
    rcurve, extenthalf, res=None,
):
    """ Get the set of points in (x, y, z) coordinates sampling the crystal
    outline
    """

    # check inputs
    if res is None:
        res = np.min(extenthalf)/5.

    # compute
    npsi = 2*int(np.ceil(extenthalf[0] / res)) + 1
    ntheta = 2*int(np.ceil(extenthalf[1] / res)) + 1

    psi, dtheta = CrystBragg_sample_outline_sphrect(
        extenthalf[0], extenthalf[1],
        npsi=npsi, ntheta=ntheta,
        include_summit=False,
    )

    vout = CrystBragg_get_noute1e2_from_psitheta(
        nout, e1, e2, psi, dtheta,
        e1e2=False, sameshape=False,
    )
    return center[:, None] + rcurve*vout


# ###############################################
#           lamb <=> bragg
# ###############################################

def get_bragg_from_lamb(lamb, d, n=None):
    """ n*lamb = 2d*sin(bragg)
    The angle bragg is defined as the angle of incidence of the emissed photon
    vector and the crystal mesh, and not the crystal dioptre.
    For record, both are parallel and coplanar when is defined parallelism
    into the crystal.
    """
    if n is None:
        n = 1
    bragg = np.full(lamb.shape, np.nan)
    sin = n*lamb/(2.*d)
    indok = np.abs(sin) <= 1.
    bragg[indok] = np.arcsin(sin[indok])
    return bragg


def get_lamb_from_bragg(bragg, d, n=None):
    """ n*lamb = 2d*sin(bragg)
    The angle bragg is defined as the angle of incidence of the emissed photon
    vector and the crystal mesh, and not the crystal dioptre.
    For record, both are parallel and coplanar when is defined parallelism
    into the crystal.
    """
    if n is None:
        n = 1
    return 2*d*np.sin(bragg) / n


# ###############################################
#           vectors <=> angles
# ###############################################


def get_vectors_from_angles(alpha, beta, nout, e1, e2):
    """Return new unit vectors according to alpha and beta entries from user
    caused by the miscut assumed on the crystal.
    """

    e1_bis = (
        np.cos(alpha)*(np.cos(beta)*e1 + np.sin(beta)*e2) - np.sin(alpha)*nout
    )

    e2_bis = np.cos(beta)*e2-np.sin(beta)*e1

    nout_bis = (
        np.cos(alpha)*nout + np.sin(alpha)*(np.cos(beta)*e1 + np.sin(beta)*e2)
    )

    nin_bis = -nout_bis

    return nin_bis, nout_bis, e1_bis, e2_bis


# ###############################################
#           Approximate solution
# ###############################################


def get_rowland_dist_from_bragg(bragg=None, rcurve=None):
    return rcurve*np.sin(bragg)


def get_approx_detector_rel(rcurve, bragg,
                            braggref=None, xiref=None,
                            bragg01=None, dist01=None,
                            tangent_to_rowland=None):
    """ Return the approximative detector position on the Rowland circle
    relatively to the Bragg crystal.
    Possibility to define tangential position of the detector to the Rowland
    circle or not.
    On WEST, the maximum miscut between two halves can be up to few
    arcmin so here, doesn't need to define the precise location of the detector
    The bragg angle is provided and naturally defined as the angle between the
    emissed photon vector and the crystal mesh.
    So, if miscut approuved, bragg is relative
    to the vector basis dmat(nout,e1,e2).
    The position of the detector, relatively to the crystal, will be so in
    another Rowland circle with its center shifted from the original one.
    """

    if tangent_to_rowland is None:
        tangent_to_rowland = True

    # distance crystal - det_center
    # bragg between incident vector and mesh
    det_dist = rcurve*np.sin(bragg)

    # det_nout and det_e1 in (nout, e1, e2) (det_e2 = e2)
    n_crystdet_rel = np.r_[-np.sin(bragg), np.cos(bragg), 0.]
    if tangent_to_rowland is True:
        bragg2 = 2.*bragg
        det_nout_rel = np.r_[-np.cos(bragg2), -np.sin(bragg2), 0.]
        det_ei_rel = np.r_[np.sin(bragg2), -np.cos(bragg2), 0.]
    else:
        det_nout_rel = -n_crystdet_rel
        det_ei_rel = np.r_[np.cos(bragg), np.sin(bragg), 0]

    # update with bragg01 and dist01
    if bragg01 is not None:
        ang = np.diff(np.sort(bragg01))
        # h = l1 tan(theta1) = l2 tan(theta2)
        # l = l2 (tan(theta1) + tan(theta2)) / tan(theta1)
        # l = l2 / cos(theta2)
        # l = l tan(theta1) / (cos(theta2) * (tan(theta1) + tan(theta2)))
        theta2 = bragg if tangent_to_rowland is True else np.pi/2
        theta1 = np.abs(bragg-bragg01[0])
        tan1 = np.tan(theta1)
        d0 = det_dist * tan1 / (np.cos(theta2) * (tan1+np.tan(theta2)))
        theta1 = np.abs(bragg-bragg01[1])
        tan1 = np.tan(theta1)
        d1 = det_dist * tan1 / (np.cos(theta2) * (tan1+np.tan(theta2)))
        if np.prod(np.sign(bragg01-bragg)) >= 0:
            d01 = np.abs(d0 - d1)
        else:
            d01 = d0 + d1
        det_dist = det_dist * dist01 / d01

    return det_dist, n_crystdet_rel, det_nout_rel, det_ei_rel


def get_det_abs_from_rel(det_dist, n_crystdet_rel, det_nout_rel, det_ei_rel,
                         summit, nout, e1, e2,
                         ddist=None, di=None, dj=None,
                         dtheta=None, dpsi=None, tilt=None):
    """ Return the absolute detector position, according to tokamak's frame,
    on the Rowland circle from its relative position to the Bragg crystal.
    If miscut approuved, bragg is relative to the vector basis
    dmat(nout,e1,e2).
    The position of the detector, relatively to the crystal, will be so in
    another Rowland circle with its center shifted from the original one.
    """

    # Reference on detector
    det_nout = (det_nout_rel[0]*nout
                + det_nout_rel[1]*e1 + det_nout_rel[2]*e2)
    det_ei = (det_ei_rel[0]*nout
                + det_ei_rel[1]*e1 + det_ei_rel[2]*e2)
    det_ej = np.cross(det_nout, det_ei)

    # Apply translation of center (ddist, di, dj)
    if ddist is None:
        ddist = 0.
    if di is None:
        di = 0.
    if dj is None:
        dj = 0.
    det_dist += ddist

    n_crystdet = (n_crystdet_rel[0]*nout
                  + n_crystdet_rel[1]*e1 + n_crystdet_rel[2]*e2)
    det_cent = summit + det_dist*n_crystdet + di*det_ei + dj*det_ej

    # Apply angles on unit vectors with respect to themselves
    if dtheta is None:
        dtheta = 0.
    if dpsi is None:
        dpsi = 0.
    if tilt is None:
        tilt = 0.

    # dtheta and dpsi
    det_nout2 = (
        (np.cos(dpsi)*det_nout + np.sin(dpsi)*det_ei)*np.cos(dtheta)
        + np.sin(dtheta)*det_ej
    )
    det_ei2 = (np.cos(dpsi)*det_ei - np.sin(dpsi)*det_nout)
    det_ej2 = np.cross(det_nout2, det_ei2)

    # tilt
    det_ei3 = np.cos(tilt)*det_ei2 + np.sin(tilt)*det_ej2
    det_ej3 = np.cross(det_nout2, det_ei3)

    return det_cent, det_nout2, det_ei3, det_ej3


# ###############################################
#           Sagital / meridional focus
# ###############################################


def calc_meridional_sagittal_focus(
    rcurve=None,
    bragg=None,
    alpha=None,
    miscut=None,
    verb=None,
):

    # Check input
    if rcurve is None or bragg is None:
        msg = (
            "Args rcurve and bragg must be provided!"
        )
        raise Exception(msg)

    verb = _check_bool(verb, vardef=True, varname='verb')
    miscut = _check_bool(
        miscut,
        vardef=_MISCUT,
        varname='miscut',
    )
    # Compute
    s_merid_ref = rcurve*np.sin(bragg)
    s_sagit_ref = -s_merid_ref/np.cos(2.*bragg)

    s_merid_unp = rcurve*(np.sin(bragg) + np.cos(bragg)*np.sin(alpha))
    s_sagit_unp = -s_merid_unp/(1-2.*np.sin(bragg+alpha)**2.)

    # verb
    if verb is True:
        mr = round(s_merid_ref, ndigits=3)
        sr = round(s_sagit_ref, ndigits=3)
        msg = (
            "Assuming a perfect crystal, from it are:\n"
            f"\t-the meridonal focus at {mr} m.\n"
            f"\t-the sagittal focus at {sr} m.\n"
        )

        if miscut is True:
            delta_merid = abs(s_merid_unp - s_merid_ref)
            delta_sagit = abs(s_sagit_unp - s_sagit_ref)
            mnp = round(s_merid_unp, ndigits=3)
            snp = round(s_sagit_unp, ndigits=3)
            mca = round(delta_merid, ndigits=3)
            sca = round(delta_sagit, ndigits=3)
            mcr = round(100. * delta_merid / s_merid_ref, ndigits=3)
            scr = round(100. * delta_sagit / s_sagit_ref, ndigits=3)
            msg += (
                f"\nConsidering a miscut angle (alpha = {alpha} rad):\n"
                f"\t-the meridonal focus at {mnp}m (delta = {mca}m / {mcr}%)\n"
                f"\t-the sagittal focus at {snp}m (delta = {sca}m / {scr}%)"
            )
        print(msg)

    return s_merid_ref, s_sagit_ref, s_merid_unp, s_sagit_unp


# ###############################################
#           Coordinates transforms
# ###############################################


def _checkformat_pts(pts=None):
    pts = np.atleast_1d(pts)
    if pts.ndim == 1:
        pts = pts.reshape((3, 1))
    if pts.shape[0] != 3 or pts.ndim < 2:
        msg = "pts must be a (3, ...) array of (X, Y, Z) coordinates!"
        raise Exception(msg)
    return pts


def checkformat_vectang(Z, nn, frame_cent, frame_ang):
    # Check / format inputs
    nn = np.atleast_1d(nn).ravel()
    assert nn.size == 3
    nn = nn / np.linalg.norm(nn)
    Z = float(Z)

    frame_cent = np.atleast_1d(frame_cent).ravel()
    assert frame_cent.size == 2
    frame_ang = float(frame_ang)

    return Z, nn, frame_cent, frame_ang


def get_e1e2_detectorplane(nn, nIn):
    e1 = np.cross(nn, nIn)
    e1n = np.linalg.norm(e1)
    if e1n < 1.e-10:
        e1 = np.array([nIn[2], -nIn[1], 0.])
        e1n = np.linalg.norm(e1)
    e1 = e1 / e1n
    e2 = np.cross(nn, e1)
    e2 = e2 / np.linalg.norm(e2)
    return e1, e2


# To be made cleaner vs option 0/ 1 => grid = True, False
def calc_xixj_from_braggphi(
    det_cent=None,
    det_nout=None, det_ei=None, det_ej=None,
    det_outline=None,
    summit=None, nout=None, e1=None, e2=None,
    bragg=None, phi=None,
    option=None, strict=None,
):
    """ Several options for shapes

    de_cent, det_nout, det_ei and det_ej are always of shape (3,)

    option:
        0:
            (summit, e1, e2).shape = (3,)
            (bragg, phi).shape = (nbragg,)
            => (xi, xj).shape = (nbragg,)
        1:
            (summit, e1, e2).shape = (3, nlamb, npts, nbragg)
            (bragg, phi).shape = (nlamb, npts, nbragg)
            => (xi, xj).shape = (nlamb, npts, nbragg)
    """
    # check inputs
    if strict is None:
        strict = True

    # Check option
    gdet = [det_cent, det_nout, det_ei, det_ej]
    g0 = [summit, nout, e1, e2]
    g1 = [bragg, phi]

    # check nbroadcastable
    _are_broadcastable(bragg=bragg, phi=phi)
    assert all([gg.shape == (3,) for gg in gdet]), "gdet no broadcast!"
    assert all([gg.shape == g0[0].shape for gg in g0]), "g0 no broadcast!"
    lc = [
        g0[0].size == 3 and g1[0].ndim == 1,
        g0[0].ndim in [4, 5] and g0[0].shape[0] == 3
        and phi.shape == g0[0].shape[1:],
    ]
    if np.sum(lc) == 0:
        lstr = [
            '\t- {}: {}'.format(kk, vv.shape)
            for kk, vv in [
                ('summit', summit), ('nout', nout), ('e1', e1), ('e2', e2),
                ('bragg', bragg), ('phi', phi),
            ]
        ]
        msg = (
            "Please provide either:\n"
            + "\t- option 0:\n"
            + "\t\t- (summit, nout, e1, e2).shape[0] = 3\n"
            + "\t\t- (bragg, phi).ndim = 1\n"
            + "\t- option 1:\n"
            + "\t\t- (summit, nout, e1, e2).ndim in [4, 5]\n"
            + "\t\t- (bragg, phi).shape[0] = 3\n\n"
            + "You provided:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)
    elif all(lc):
        msg = ("Multiple options!")
        raise Exception(msg)

    if option is None:
        option = lc.index(True)
    assert (lc[0] and option == 0) or (lc[1] and option == 1)

    if option == 0:
        summit = summit.ravel()
        nout, e1, e2 = nout.ravel(), e1.ravel(), e2.ravel()
        det_cent = det_cent[:, None]
        det_nout = det_nout[:, None]
        det_ei, det_ej = det_ei[:, None], det_ej[:, None]
        summit, nout = summit[:, None], nout[:, None],
        e1, e2 = e1[:, None], e2[:, None]
    else:
        det_cent = det_cent[:, None, None, None]
        det_nout = det_nout[:, None, None, None]
        det_ei = det_ei[:, None, None, None]
        det_ej = det_ej[:, None, None, None]
        if g0[0].ndim == 5:
            det_cent = det_cent[..., None]
            det_nout = det_nout[..., None]
            det_ei = det_ei[..., None]
            det_ej = det_ej[..., None]

    # Not necessary for broadcasting (last dims first)
    # bragg = bragg[None, ...]
    # phi = phi[None, ...]

    # Compute
    vect = (
        -np.sin(bragg)*nout
        + np.cos(bragg)*(np.cos(phi)*e1 + np.sin(phi)*e2)
    )
    k = np.sum(
        (det_cent-summit)*det_nout, axis=0
        ) / np.sum(vect*det_nout, axis=0)
    pts = summit + k[None, ...]*vect
    xi = np.sum((pts - det_cent)*det_ei, axis=0)
    xj = np.sum((pts - det_cent)*det_ej, axis=0)

    # Optional: eliminate points outside the det outline
    if det_outline is not None and strict is True:
        ind = (
            (xi < np.min(det_outline[0, :]))
            | (xi > np.max(det_outline[0, :]))
            | (xj < np.min(det_outline[1, :]))
            | (xj > np.max(det_outline[1, :]))
        )
        xi[ind] = np.nan
        xj[ind] = np.nan
    return xi, xj, strict


def _calc_braggphi_from_pts_summits(
    pts=None,
    summits=None,
    vin=None, ve1=None, ve2=None,
):
    # check inputs
    _are_broadcastable(pts=pts, summits=summits, vin=vin, ve1=ve1, ve2=ve2)

    if pts.ndim != summits.ndim:
        msg = "Differents dimensions!"
        raise Exception(msg)

    # compute
    vect = pts - summits
    vect = vect / np.sqrt(np.sum(vect**2, axis=0))[None, ...]
    bragg = np.arcsin(np.sum(vect*vin, axis=0))
    if np.any(bragg < 0.):
        msg = (
            "There seems to be negative bragg angles!\n"
            + "  => double-check inputs!"
        )
        raise Exception(msg)
    phi = np.arctan2(np.sum(vect*ve2, axis=0), np.sum(vect*ve1, axis=0))
    return bragg, phi


def calc_braggphi_from_xixjpts(
    pts=None,
    xi=None, xj=None, det=None,
    summit=None, nin=None, e1=None, e2=None,
    grid=None,
):
    """ Return bragg phi for pts or (xj, xi) seen from (summit, nin, e1, e2)

    Either provide:
        pts => (3, npts)
        xi, xj => pts with shape (3, nxi, nxj)

    summit, nin, e1, e2 must have the same shape (3, nsumm)

    bragg.shape = (nsum, )

    if grid is True:
        all pts evaluated for all summ/nin
        return  (nsumm, npts) or (nsum, nxi, nxj) arrays
    else:
        each pts has a unique corresponding summ/nin (except possibly ndtheta)
        return (npts,) or (nxi, nxj) arrays
            or (npts, ndtheta) or (nxi, nxj, ndtheta) arrays
    """

    # --------------
    # Check format
    if grid is None:
        grid = True

    # Either pts or (xi, xj)
    lc = [pts is not None, all([xx is not None for xx in [xi, xj]])]
    if np.sum(lc) != 1:
        msg = "Provide either pts xor (xi, xj)!"
        raise Exception(msg)

    if lc[0]:
        # pts
        pts = _checkformat_pts(pts)

    elif lc[1]:

        # xi, xj => compute pts using det
        c0 = (
            isinstance(det, dict)
            and all([
                ss in det.keys() and len(det[ss]) == 3
                for ss in ['cent', 'ei', 'ej']
            ])
        )
        if not c0:
            msg = (
                "Arg det must be provided as a dict if xi, xj are provided!\n"
                + "Provided: {}".format(det)
            )
            raise Exception(msg)

        # pts from xi, xj
        xi, xj, (xii, xjj) = _checkformat_xixj(xi, xj)
        if xii.shape != xjj.shape:
            msg = "xi and xj must have the same shape!"
            raise Exception(msg)
        assert xii.ndim in [1, 2]

        if xii.ndim == 1:
            pts = (det['cent'][:, None]
                   + xii[None, :]*det['ei'][:, None]
                   + xjj[None, :]*det['ej'][:, None])
        else:
            pts = (det['cent'][:, None, None]
                   + xii[None, ...]*det['ei'][:, None, None]
                   + xjj[None, ...]*det['ej'][:, None, None])

    c0 = summit.shape == nin.shape == e1.shape == e2.shape
    if not c0:
        msg = "(summit, nin, e1, e2) must all have the same shape"
        raise Exception(msg)

    c0 = (
        (
            grid is True
            and pts.ndim in [1, 2, 3]
            and summit.ndim in [1, 2, 3, 4, 5]
        )
        or (
            grid is False
            and pts.ndim == summit.ndim
        )
    )
    if not c0:
        msg = (
            "Args pts and summit/nin/e1/e2 must be such that:\n"
            + "\t- grid = True:\n"
            + "\t\tpts.ndim in [1, 2, 3] and pts.shape[0] == 3\n"
            + "\t\tsummit.ndim in [1, 2, 3, 4, 5]\n"
            + "\t- grid = False:\n"
            + "\t\tpts can be directly broadcasted to summit (and same dim)\n"
            + "  You provided:\n"
            + "\t- grid: {}\n".format(grid)
            + "\t- pts.shape = {}\n".format(pts.shape)
            + "\t- summit.shape = {}".format(summit.shape)
        )
        raise Exception(msg)

    # --------------
    # Prepare
    # This part should be re-checked for all combinations!
    if grid is False:
        return _calc_braggphi_from_pts_summits(
            pts=pts,
            summits=summit,
            vin=nin, ve1=e1, ve2=e2,
        )

    else:
        # Typical dim
        # (3, npts, nlamb, ndtheta, 2)
        # (3, nxi, nxj, nlamb, ndtheta, 2)
        ptsdim = pts.ndim
        sumdim = summit.ndim
        if ptsdim == 2:
            if sumdim == 1:
                summit = summit[:, None]
                nin = nin[:, None]
                e1, e2 = e1[:, None], e2[:, None]
            else:
                summit = summit[:, None, ...]
                nin = nin[:, None, ...]
                e1, e2 = e1[:, None, ...], e2[:, None, ...]
        else:
            if sumdim == 1:
                summit = summit[:, None, None]
                nin = nin[:, None, None]
                e1, e2 = e1[:, None, None], e2[:, None, None]
            else:
                summit = summit[:, None, None, ...]
                nin = nin[:, None, None, ...]
                e1, e2 = e1[:, None, None, ...], e2[:, None, None, ...]
        if sumdim in [1, 2]:
            pts = pts[..., None]
        elif sumdim == 3:
            pts = pts[..., None, None]
        elif sumdim == 4:
            pts = pts[..., None, None, None]

        # --------------
        # Compute
        # Everything has shape (3, nxi0, nxi1, npts0, npts1) => sum on axis=0
        # or is broadcastable
        return _calc_braggphi_from_pts_summits(
            pts=pts,
            summits=summit,
            vin=nin, ve1=e1, ve2=e2,
        )


# ###############################################
#           lamb available from pts
# ###############################################


def _get_lamb_avail_from_pts_phidtheta_xixj(
    cryst=None,
    lamb=None,
    n=None,
    ndtheta=None,
    pts=None,
    miscut=None,
    return_phidtheta=None,
    return_xixj=None,
    strict=None,
    det=None,
):
    """

    Inputs
    ------
        pts = (3, npts) array
        lamb = (npts, nlamb) array

    Return
    ------
        lamb    (npts, nlamb)
        xi      (npts, nlamb, ndtheta, 2)   There can be 2 solutions
        xj      (npts, nlamb, ndtheta, 2)
        phi     (npts, nlamb, ndtheta, 2)
        dtheta  (npts, nlamb, ndtheta, 2)

    """

    keepon = return_phidtheta or return_xixj or strict

    if keepon:
        # Continue to get phi, dtheta...
        bragg = cryst._checkformat_bragglamb(lamb=lamb, n=n)

        # Compute dtheta / psi for each pts and lamb
        npts, nlamb = lamb.shape
        dtheta = np.full((npts, nlamb, ndtheta, 2), np.nan)
        psi = np.full((npts, nlamb, ndtheta, 2), np.nan)
        phi = np.full((npts, nlamb, ndtheta, 2), np.nan)
        for ii in range(nlamb):
            (
                dtheta[:, ii, :, :], psi[:, ii, :, :],
                phi[:, ii, :, :],
            ) = cryst._calc_dthetapsiphi_from_lambpts(
                pts=pts, bragg=bragg[:, ii], lamb=None,
                n=n, ndtheta=ndtheta,
                miscut=miscut,
                grid=False,
            )[:3]

        if return_xixj is True or strict is True:
            xi, xj, strict = cryst.calc_xixj_from_braggphi(
                phi=phi + np.pi,    # from plasma to det
                bragg=bragg[..., None, None],
                n=n,
                dtheta=dtheta,
                psi=psi,
                det=det,
                data=None,
                miscut=miscut,
                strict=strict,
                return_strict=True,
                plot=False,
                dax=None,
            )
            if strict is True and np.any(np.isnan(xi)):
                indnan = np.isnan(xi)
                lamb[np.all(np.all(indnan, axis=-1), axis=-1)] = np.nan
                if return_phidtheta:
                    phi[indnan] = np.nan
                    dtheta[indnan] = np.nan
                    psi[indnan] = np.nan

    # -----------
    # return

    if return_phidtheta and return_xixj:
        return lamb, phi, dtheta, psi, xi, xj
    elif return_phidtheta:
        return lamb, phi, dtheta, psi
    elif return_xixj:
        return lamb, xi, xj
    else:
        return lamb


# ###############################################
#           2D spectra to 1D
# ###############################################


def get_lambphifit(lamb, phi, nxi, nxj):
    lambD = np.nanmax(lamb)-np.nanmin(lamb)
    lambfit = np.nanmin(lamb) + lambD*np.linspace(0, 1, nxi)
    phiD = np.nanmax(phi) - np.nanmin(phi)
    phifit = np.nanmin(phi) + phiD*np.linspace(0, 1, nxj)
    return lambfit, phifit


def _calc_spect1d_from_data2d(ldata, lamb, phi,
                              nlambfit=None, nphifit=None,
                              spect1d=None, mask=None,
                              vertsum1d=None):
    # Check / format inputs
    if spect1d is None:
        spect1d = 'mean'
    if isinstance(ldata, np.ndarray):
        ldata = [ldata]
    lc = [isinstance(spect1d, tuple) and len(spect1d) == 2,
          (isinstance(spect1d, list)
           and all([isinstance(ss, tuple) and len(ss) == 2
                    for ss in spect1d])),
          spect1d in ['mean', 'cent']]
    if lc[0]:
        spect1d = [spect1d]
    elif lc[1]:
        pass
    elif lc[2]:
        if spect1d == 'cent':
            spect1d = [(0., 0.2)]
            nspect = 1
    else:
        msg = ("spect1d must be either:\n"
               + "\t- 'mean': the avearge spectrum\n"
               + "\t- 'cent': the central spectrum +/- 20%\n"
               + "\t- (target, tol); a tuple of 2 floats:\n"
               + "\t\ttarget: the central value of the window in [-1,1]\n"
               + "\t\ttol:    the window tolerance (width) in [0,1]\n"
               + "\t- list of (target, tol)")
        raise Exception(msg)

    if not isinstance(nlambfit, int) or not isinstance(nphifit, int):
        msg = ("nlambfit and nphifit must be int!\n"
               + "\t- nlambfit provided: {}\n".format(nlambfit)
               + "\t- nphifit provided : {}\n".format(nphifit))
        raise Exception(msg)

    if vertsum1d is None:
        vertsum1d = True

    # Compute lambfit / phifit and spectrum1d
    if mask is not None:
        for ii in range(len(ldata)):
            ldata[ii][~mask] = np.nan
    lambfit, phifit = get_lambphifit(lamb, phi, nlambfit, nphifit)
    lambfitbins = 0.5*(lambfit[1:] + lambfit[:-1])
    ind = np.digitize(lamb, lambfitbins)

    # Get phi window
    if spect1d == 'mean':
        phiminmax = np.r_[phifit.min(), phifit.max()][None, :]
        spect1d_out = [np.array([np.nanmean(dd[ind == jj])
                                 for jj in np.unique(ind)])[None, :]
                       for dd in ldata]
    else:
        nspect = len(spect1d)
        dphi = np.nanmax(phifit) - np.nanmin(phifit)
        spect1d_out = [np.full((nspect, lambfit.size), np.nan)
                       for dd in ldata]
        phiminmax = np.full((nspect, 2), np.nan)
        for ii in range(nspect):
            phicent = np.nanmean(phifit) + spect1d[ii][0]*dphi/2.
            indphi = np.abs(phi - phicent) < spect1d[ii][1]*dphi
            for jj in np.unique(ind):
                indj = indphi & (ind == jj)
                if np.any(indj):
                    for ij in range(len(ldata)):
                        spect1d_out[ij][ii, jj] = np.nanmean(ldata[ij][indj])
            phiminmax[ii, :] = (np.nanmin(phi[indphi]),
                                np.nanmax(phi[indphi]))

    if vertsum1d is True:
        phifitbins = 0.5*(phifit[1:] + phifit[:-1])
        ind = np.digitize(phi, phifitbins)
        vertsum1d = [np.array([np.nanmean(dd[ind == ii])
                               for ii in np.unique(ind)])
                     for dd in ldata]
    if len(ldata) == 1:
        spect1d_out = spect1d_out[0]
        if vertsum1d is not False:
            vertsum1d = vertsum1d[0]
    return spect1d_out, lambfit, phifit, vertsum1d, phiminmax


# ###############################################
#           From plasma pts
# ###############################################


def calc_dthetapsiphi_from_lambpts(
    pts,
    bragg,
    summit=None, rcurve=None,
    nout=None, e1=None, e2=None,
    extenthalf=None,
    ndtheta=None,
    grid=None,
):
    """ Return (dtheta, psi) of pts on crystal where bragg diffraction happens

    For given pts and lamb/bragg

    For each pts/lamb, there may be up to 2 arcs on the crystal
    Only returns valid solution (inside extenthalf), with nan elsewhere

    psi and dtheta returned as (nlamb, npts, 2, ndtheta) arrays

    Here nout, e1, e2 are at the unique crystal summit!

    """

    # Check input
    if ndtheta is None:
        ndtheta = 10

    npts = pts.shape[1]
    nlamb = bragg.size
    if grid is None:
        grid = True
    if grid is False:
        if nlamb != npts:
            msg = "If grid = False, lamb.shape should be (pts.shape[1],)"
            raise Exception(msg)

    # Prepare output
    if grid is True:
        scaPCem = np.full((nlamb, npts, 2), np.nan)
        dtheta = np.full((nlamb, npts, ndtheta, 2), np.nan)
        psi = np.full((nlamb, npts, ndtheta, 2), np.nan)
        num = np.full((nlamb, npts, ndtheta, 2), np.nan)
        angextra = np.full((nlamb, npts, ndtheta, 2), np.nan)
        dtheta_u = np.full((nlamb, npts, ndtheta), np.nan)
        psi_u = np.full((nlamb, npts, ndtheta), np.nan)
        sol1 = np.full((nlamb, npts), np.nan)
        sol2 = np.full((nlamb, npts), np.nan)
    else:
        scaPCem = np.full((npts, 2), np.nan)
        dtheta = np.full((npts, ndtheta, 2), np.nan)
        psi = np.full((npts, ndtheta, 2), np.nan)
        num = np.full((npts, ndtheta, 2), np.nan)
        angextra = np.full((npts, ndtheta, 2), np.nan)
        dtheta_u = np.full((npts, ndtheta), np.nan)
        psi_u = np.full((npts, ndtheta), np.nan)
        sol1 = np.full((npts,), np.nan)
        sol2 = np.full((npts,), np.nan)

    # Get to scalar product scaPCem
    # Already ok for miscut (via nout)
    center = summit - rcurve*nout
    PC = center[:, None] - pts
    PCnorm2 = np.sum(PC**2, axis=0)
    cos2 = np.cos(bragg)**2
    # PM.CM = Rsca + R**2  (ok)
    # PMCM = PMnR*sin       (ok)
    # PMn2 = PCn2*sin2 + 2Rsin2*sca + R2sin2
    #
    # sca**2 + 2Rcos2*sca + R2cos2 - PCnsin2 = 0
    if grid is True:
        deltaon4 = np.sin(bragg)[:, None]**2*(
            PCnorm2[None, :] - rcurve**2*cos2[:, None]
        )
    else:
        deltaon4 = np.sin(bragg)**2 * (PCnorm2 - rcurve**2*cos2)

    # Get two relevant solutions
    ind = deltaon4 >= 0.
    if grid is True:
        cos2 = np.repeat(cos2[:, None], npts, axis=1)[ind]
        PCnorm = np.tile(np.sqrt(PCnorm2), (nlamb, 1))[ind]
    else:
        cos2 = cos2[ind]
        PCnorm = np.sqrt(PCnorm2)[ind]
    sol1 = -rcurve*cos2 - np.sqrt(deltaon4[ind])
    sol2 = -rcurve*cos2 + np.sqrt(deltaon4[ind])

    # Only keep solution going outward sphere
    # scaPMem = scaPCem + rcurve >= 0
    ind1 = (sol1 >= -rcurve)
    ind2 = (sol2 >= -rcurve)

    sol1 = sol1[ind1]
    sol2 = sol2[ind2]
    if grid is True:
        indn = ind.nonzero()
        ind1 = [indn[0][ind1], indn[1][ind1]]
        ind2 = [indn[0][ind2], indn[1][ind2]]
        scaPCem[ind1[0], ind1[1], 0] = sol1
        scaPCem[ind2[0], ind2[1], 1] = sol2
    else:
        indn = ind.nonzero()[0]
        scaPCem[indn[ind1], 0] = sol1
        scaPCem[indn[ind2], 1] = sol2
    ind = ~np.isnan(scaPCem)

    # Get equation on PCem
    # CM = rcurve * (sin(dtheta)e2 + cos(dtheta)(cos(psi)nout + sin(psi)e1))
    # PC.eM = scaPCem, thus introducing Z = PC.e2, Y = PC.e1, X = PC.nout
    # Xcos(dtheta)cos(psi) + Ycos(dtheta)sin(psi) + Zsin(dtheta) = scaPCem
    # dtheta is specified, psi is deduced
    X = np.sum(PC*nout[:, None], axis=0)
    Y = np.sum(PC*e1[:, None], axis=0)
    Z = np.sum(PC*e2[:, None], axis=0)

    if grid is True:
        scaPCem = np.repeat(scaPCem[:, :, None, :], ndtheta, axis=2)
    else:
        scaPCem = np.repeat(scaPCem[:, None, :], ndtheta, axis=1)

    # broadcast and specify dtheta
    ind = ~np.isnan(scaPCem)
    if grid is True:
        XYnorm = np.repeat(
            np.repeat(
                np.repeat(
                    np.sqrt(X**2 + Y**2)[None, :], nlamb, axis=0,
                )[..., None],
                ndtheta,
                axis=-1,
            )[..., None],
            2,
            axis=-1,
        )[ind]
        Z = np.repeat(
            np.repeat(
                np.repeat(Z[None, :], nlamb, axis=0)[..., None],
                ndtheta,
                axis=-1,
            )[..., None],
            2,
            axis=-1,
        )[ind]
        # Define angextra to get
        # sin(psi + angextra) = (scaPCem - Z*sin(theta)) / (XYnorm*cos(theta))
        angextra[ind] = np.repeat(
            np.repeat(
                np.repeat(np.arctan2(X, Y)[None, :], nlamb, axis=0)[..., None],
                ndtheta, axis=-1)[..., None],
            2, axis=-1)[ind]
        dtheta[ind] = np.repeat(
            np.repeat(
                np.repeat(
                    extenthalf[1]*np.linspace(-1, 1, ndtheta)[:, None],
                    2, axis=1)[None, ...],
                npts, axis=0)[None, ...],
            nlamb, axis=0)[ind]
    else:
        XYnorm = np.repeat(
            np.repeat(np.sqrt(X**2+Y**2)[:, None], ndtheta, axis=1)[..., None],
            2,
            axis=-1,
        )[ind]
        Z = np.repeat(
            np.repeat(Z[:, None], ndtheta, axis=1)[..., None],
            2,
            axis=-1,
        )[ind]
        angextra[ind] = np.repeat(
            np.repeat(np.arctan2(X, Y)[:, None], ndtheta, axis=-1)[..., None],
            2,
            axis=-1,
        )[ind]
        dtheta[ind] = np.repeat(
            np.repeat(
                extenthalf[1]*np.linspace(-1, 1, ndtheta)[:, None],
                2,
                axis=1,
            )[None, ...],
            npts,
            axis=0,
        )[ind]

    num[ind] = (
        (scaPCem[ind] - Z*np.sin(dtheta[ind])) / (XYnorm*np.cos(dtheta[ind]))
    )
    ind[ind] = np.abs(num[ind]) <= 1.
    # minus psi ref ?
    # value of psi + angextra with cos > 0
    psiang_pos = np.arcsin(num[ind])
    psi1 = np.full(psi.shape, np.nan)
    psi2 = np.full(psi.shape, np.nan)
    psi1[ind] = (np.arctan2(num[ind], np.cos(psiang_pos)) - angextra[ind])
    psi2[ind] = (np.arctan2(num[ind], -np.cos(psiang_pos)) - angextra[ind])

    # Make sure only one of the 2 psi is correct
    ind1 = np.copy(ind)
    ind2 = np.copy(ind)
    ind1[ind] = np.abs(psi1[ind]) <= extenthalf[0]
    ind2[ind] = np.abs(psi2[ind]) <= extenthalf[0]
    assert not np.any(ind1 & ind2), "Multiple solutions for psi!"
    ind = ind1 | ind2

    # Finally store into psi
    psi[ind1] = psi1[ind1]
    psi[ind2] = psi2[ind2]
    psi[~ind] = np.nan
    dtheta[~ind] = np.nan
    if np.any(np.sum(ind, axis=-1) == 2):
        msg = (
            "\nDouble solutions found for {} / {} points!".format(
                np.sum(np.sum(ind, axis=-1) == 2),
                np.prod(ind.shape[:-1]),
            )
        )
        warnings.warn(msg)

    return dtheta, psi, ind, grid


"""
This module is the computational part of the geometrical module of ToFu
"""

# Built-in
import os
import warnings
from xml.dom import minidom

# Common
import numpy as np
import scipy.interpolate as scpinterp
import scipy.integrate as scpintg
import matplotlib.colors as mplcol
from inspect import signature as insp


# ToFu-specific
try:
    import tofu.geom._def as _def
    import tofu.geom._GG as _GG
except Exception:
    from . import _def as _def
    from . import _GG as _GG


_LTYPES = [int, float, np.int_, np.float64]
_RES = 0.1


###############################################################################
#                            Default parameters
###############################################################################


_SAMPLE_RES = {
    'edge': 0.02,
    'cross': 0.1,
    'surface': 0.1,
    'volume': 0.1,
}
_SAMPLE_RESMODE = {
    'edge': 'abs',
    'cross': 'abs',
    'surface': 'abs',
    'volume': 'abs',
}


def _check_float(var=None, varname=None, vardef=None):
    if var is None:
        var = vardef
    if not type(var) in _LTYPES:
        msg = (
            "Arg {} must be a float!\n".format(varname)
            + "Provided: {}".format(type(var))
        )
        raise Exception(msg)
    return var

###############################################################################
#                            Ves functions
###############################################################################


# ==============================================================================
# Interfacing functions
# ==============================================================================


def _get_pts_from_path_svg(
    path_str=None,
    res=None,
    k0=None,
):

    # Check inputs
    res = _check_float(var=res, varname='res', vardef=_RES)

    # try loading
    try:
        from svg.path import parse_path
    except Exception as err:
        msg = (
            str(err)
            + "\n\nYou do not seem to have svg.path installed\n"
            + "It is an optional dependency only used for this method\n"
            + "To use from_svg(), please install svg.path using:\n"
            + "\tpip install svg.path"
        )
        raise Exception(msg)

    lpath = parse_path(path_str)

    lpath._calc_lengths()
    fract = lpath._fractions

    pos = []
    for ii, pat in enumerate(lpath):
        if pat.__class__.__name__ == 'Line':
            pos.append(np.r_[fract[ii]])
        elif pat.__class__.__name__ == 'Move':
            pos.append(np.r_[fract[ii]])
        elif pat.__class__.__name__ == 'Close':
            pos.append(np.r_[fract[ii]])
        else:
            npts = int(np.ceil(pat.length() / res))
            pos.append(
                np.linspace(fract[ii], fract[ii+1], npts, endpoint=False)
            )

    pos = np.unique(np.concatenate(pos))
    ind1 = np.abs(pos-1.) < 1e-14
    if np.sum(ind1) == 1:
        pos[ind1] = 1.
    elif np.sum(ind1) > 1:
        msg = "Several 1!"
        raise Exception(msg)
    pts = np.array([lpath.point(po) for po in pos])
    pts = np.array([pts.real, pts.imag])

    # Check for reference line
    isref = False
    if 'z' not in path_str.lower():
        if pts.shape[1] == 2:
            isref = True
        elif np.allclose(pts[:, 0], pts[:, -1]):
            pass
        else:
            pts = np.concatenate((pts, pts[:, 0:1]), axis=1)
            msg = (
                f"Non-conform path '{k0}' identified!\n"
                "All path must be either:\n"
                "\t- closed\n"
                "\t- or a unique straight line with 2 points\n"
                "Provided:\n"
                f"path_str: {path_str}\n"
                f"  => closed automatically"
            )
            warnings.warn(msg)

    return pts, isref


def get_paths_from_svg(
    pfe=None,
    res=None,
    r0=None,
    z0=None,
    point_ref1=None,
    point_ref2=None,
    length_ref=None,
    scale=None,
    verb=None,
):

    # check input
    c0 = isinstance(pfe, str) and os.path.isfile(pfe) and pfe.endswith('.svg')
    if not c0:
        msg = (
            "Arg pfe should be a path to a valid .svg file!\n"
            + "Provided:\n\t{}".format(pfe)
        )
        raise Exception(msg)
    pfe = os.path.abspath(pfe)

    # r0, z0, scale
    z0 = _check_float(var=z0, varname='z0', vardef=0.)
    r0 = _check_float(var=r0, varname='r0', vardef=0.)
    scale = _check_float(var=scale, varname='scale', vardef=1.)

    # verb
    if verb is None:
        verb = True
    if not isinstance(verb, bool):
        msg = (
            "Arg verb must be a bool!\n"
            + "Provided:\n\t{}".format(verb)
        )
        raise Exception(msg)

    # Predefine useful var
    doc = minidom.parse(pfe)

    # Try extract raw data
    try:
        dpath = {
            path.getAttribute('id').replace('\n', '').replace('""', ''): {
                'poly': path.getAttribute('d'),
                'color': path.getAttribute('style')
            }
            for path in doc.getElementsByTagName('path')
        }
    except Exception as err:
        msg = (
            "Could not extract path coordinates from {}".format(pfe)
        )
        raise Exception(msg)

    # Derive usable data
    kstr = 'fill:'
    dpath = {k0: v0 for k0, v0 in dpath.items() if kstr in v0['color']}
    lk = list(dpath.keys())
    ref = None
    for ii, k0 in enumerate(lk):

        v0 = dpath[k0]
        poly, isref = _get_pts_from_path_svg(v0['poly'], res=res, k0=k0)
        if isref is True:
            ref = poly
            del dpath[k0]
            continue
        dpath[k0]['poly'] = poly

        # class and color
        color = v0['color'][v0['color'].index(kstr) + len(kstr):].split(';')[0]
        if color == 'none':
            dpath[k0]['cls'] = 'Ves'
            color = None
        else:
            if mplcol.to_rgb(color) == (1., 0., 0.):
                dpath[k0]['cls'] = 'CoilPF'
            else:
                dpath[k0]['cls'] = 'PFC'
        dpath[k0]['color'] = color

    # Check for negative r
    lkneg = [k0 for k0, v0 in dpath.items() if np.any(v0['poly'][0, :] <= 0.)]
    if len(lkneg) > 0.:
        lstr = ['\t- {}'.format(k0) for k0 in lkneg]
        msg = (
            "With the chosen r0 ({}) some structure have negative r values\n"
            + "This is impossible in a toroidal coordinate system\n"
            + "  => the following structures are removed:\n"
            + "\n".join(lstr)
        )
        if len(lkneg) == len(dpath):
            raise Exception(msg)
        else:
            warnings.warn(msg)
        dpath = {k0: dpath[k0] for k0 in dpath.keys() if k0 not in lkneg}

    # Set origin and rescale
    if ref is not None:
        lc = [
            point_ref1 is not None and point_ref2 is not None,
            point_ref1 is not None and length_ref is not None,
        ]
        if not any(lc):
            msg = (
                "Arg reference line for scaling has been detected!\n"
                + "But it cannot be used without providing:\n"
                + "\t- point_ref1 + point_ref2: iterables of len() = 2\n"
                + "\t- point_Ref1 + length_ref: iterable len() = 2 + scalar\n"
            )
            warnings.warn(msg)
        else:
            unit = np.diff(ref, axis=1)
            unit = unit / np.linalg.norm(unit)
            unit = np.array([[unit[0, 0]], [-unit[1, 0]]])
            if not lc[0]:
                point_ref2 = np.array(point_ref1)[:, None] + length_ref*unit

            # if horizontal (resp. vertical line) => coef = inf
            # => assume equal scale for r and z instead to avoid inf
            eps = 1.e-8
            if np.abs(unit[0, 0]) > eps:
                r_coef = (
                    (point_ref2[0]-point_ref1[0]) / (ref[0, 1] - ref[0, 0])
                )
            if np.abs(unit[1, 0]) > eps:
                z_coef = (
                    (point_ref2[1]-point_ref1[1]) / (ref[1, 1] - ref[1, 0])
                )
            if np.abs(unit[0, 0]) < eps:
                # vertical line => assume rscale  = zscale
                r_coef = -z_coef
            if np.abs(unit[1, 0]) < eps:
                # horizontal line => assume rscale  = zscale
                z_coef = -r_coef
            r_offset = point_ref1[0] - r_coef*ref[0, 0]
            z_offset = point_ref1[1] - z_coef*ref[1, 0]

            for k0 in dpath.keys():
                dpath[k0]['poly'] = np.array([
                    r_coef*dpath[k0]['poly'][0, :] + r_offset,
                    z_coef*dpath[k0]['poly'][1, :] + z_offset,
                ])
    else:
        for k0 in dpath.keys():
            dpath[k0]['poly'] = np.array([
                scale*(dpath[k0]['poly'][0, :] - r0),
                scale*(-dpath[k0]['poly'][1, :] - z0),
            ])

    # verb
    if verb is True:
        lVes = sorted([k0 for k0, v0 in dpath.items() if v0['cls'] == 'Ves'])
        lPFC = sorted([k0 for k0, v0 in dpath.items() if v0['cls'] == 'PFC'])
        lCoilPF = sorted([
            k0 for k0, v0 in dpath.items() if v0['cls'] == 'CoilPF'
        ])
        lobj = [
            '\t- {}: {} ({} pts, {})'.format(
                dpath[k0]['cls'], k0,
                dpath[k0]['poly'].shape[1], dpath[k0]['color'],
            )
            for k0 in lVes + lPFC + lCoilPF
        ]
        msg = (
            "The following structures were loaded:\n".format(pfe)
            + "\n".join(lobj)
            + "\nfrom {}".format(pfe)
        )
        print(msg)

    return dpath


# ==============================================================================
# = Ves sub-functions
# ==============================================================================


def _Struct_set_Poly(
    Poly, pos=None, extent=None, arrayorder="C", Type="Tor", Clock=False
):
    """ Compute geometrical attributes of a Struct object """

    # Make Poly closed, counter-clockwise, with '(cc,N)' layout and arrayorder
    try:
        Poly = _GG.format_poly(Poly, order="C", Clock=False, close=True,
                               Test=True)
    except Exception as excp:
        print(excp)
    assert Poly.shape[0] == 2, "Arg Poly must be a 2D polygon !"
    fPfmt = np.ascontiguousarray if arrayorder == "C" else np.asfortranarray

    # Get all remarkable points and moments
    NP = Poly.shape[1] - 1
    P1Max = Poly[:, np.argmax(Poly[0, :])]
    P1Min = Poly[:, np.argmin(Poly[0, :])]
    P2Max = Poly[:, np.argmax(Poly[1, :])]
    P2Min = Poly[:, np.argmin(Poly[1, :])]
    BaryP = np.sum(Poly[:, :-1], axis=1, keepdims=False) / (Poly.shape[1] - 1)
    BaryL = np.array(
        [(P1Max[0] + P1Min[0]) / 2.0, (P2Max[1] + P2Min[1]) / 2.0]
    )
    BaryS, Surf = _GG.poly_area_and_barycenter(Poly, NP)

    # Get lim-related indicators
    noccur = int(pos.size)
    Multi = noccur > 1

    # Get Tor-related quantities
    if Type.lower() == "lin":
        Vol, BaryV = None, None
    else:
        Vol, BaryV = _GG.Poly_VolAngTor(Poly)
        if Vol <= 0.0:
            msg = ("Pb. with volume computation for Struct of type 'Tor' !\n"
                   + "\t- Vol = {}\n".format(Vol)
                   + "\t- Poly = {}\n\n".format(str(Poly))
                   + "  => Probably corrupted polygon\n"
                   + "  => Please check polygon is not self-intersecting")
            raise Exception(msg)

    # Compute the non-normalized vector of each side of the Poly
    Vect = np.diff(Poly, n=1, axis=1)
    Vect = fPfmt(Vect)

    # Compute the normalised vectors directed inwards
    Vin = np.array([Vect[1, :], -Vect[0, :]])
    Vin = -Vin  # Poly is Counter Clock-wise as defined above
    Vin = Vin / np.hypot(Vin[0, :], Vin[1, :])[np.newaxis, :]
    Vin = fPfmt(Vin)

    poly = _GG.format_poly(
        Poly,
        order=arrayorder,
        Clock=Clock,
        close=False,
        Test=True,
    )

    # Get bounding circle
    circC = BaryS
    r = np.sqrt(np.sum((poly - circC[:, np.newaxis]) ** 2, axis=0))
    circr = np.max(r)

    dout = {
        "Poly": poly,
        "pos": pos,
        "extent": extent,
        "noccur": noccur,
        "Multi": Multi,
        "nP": NP,
        "P1Max": P1Max,
        "P1Min": P1Min,
        "P2Max": P2Max,
        "P2Min": P2Min,
        "BaryP": BaryP,
        "BaryL": BaryL,
        "BaryS": BaryS,
        "BaryV": BaryV,
        "Surf": Surf,
        "VolAng": Vol,
        "Vect": Vect,
        "VIn": Vin,
        "circ-C": circC,
        "circ-r": circr,
        "Clock": Clock,
    }
    return dout


def _Ves_get_InsideConvexPoly(
    Poly,
    P2Min,
    P2Max,
    BaryS,
    RelOff=_def.TorRelOff,
    ZLim="Def",
    Spline=True,
    Splprms=_def.TorSplprms,
    NP=_def.TorInsideNP,
    Plot=False,
    Test=True,
):
    if Test:
        assert type(RelOff) is float, "Arg RelOff must be a float"
        assert (
            ZLim is None or ZLim == "Def" or type(ZLim) in [tuple, list]
        ), "Arg ZLim must be a tuple (ZlimMin, ZLimMax)"
        assert type(Spline) is bool, "Arg Spline must be a bool !"
    if ZLim is not None:
        if ZLim == "Def":
            ZLim = (
                P2Min[1] + 0.1 * (P2Max[1] - P2Min[1]),
                P2Max[1] - 0.05 * (P2Max[1] - P2Min[1]),
            )
        indZLim = (Poly[1, :] < ZLim[0]) | (Poly[1, :] > ZLim[1])
        if Poly.shape[1] - indZLim.sum() < 10:
            msg = "Poly seems to be Convex and simple enough !"
            msg += "\n  Poly.shape[1] - indZLim.sum() < 10"
            warnings.warn(msg)
            return Poly
        Poly = np.delete(Poly, indZLim.nonzero()[0], axis=1)
    if np.all(Poly[:, 0] == Poly[:, -1]):
        Poly = Poly[:, :-1]
    Np = Poly.shape[1]
    if Spline:
        BarySbis = np.tile(BaryS, (Np, 1)).T
        Ptemp = (1.0 - RelOff) * (Poly - BarySbis)
        # Poly = BarySbis + Ptemp
        Ang = np.arctan2(Ptemp[1, :], Ptemp[0, :])
        Ang, ind = np.unique(Ang, return_index=True)
        Ptemp = Ptemp[:, ind]
        # spline parameters
        ww = Splprms[0] * np.ones((Np + 1,))
        ss = Splprms[1] * (Np + 1)  # smoothness parameter
        kk = Splprms[2]  # spline order
        nest = int(
            (Np + 1) / 2.0
        )  # estimate of number of knots needed (-1 = maximal)
        # Find the knot points

        # TODO @DV : we can probably get rid of this
        # tckp,uu = scpinterp.splprep([np.append(Ptemp[0,:],Ptemp[0,0]),
        # np.append(Ptemp[1,:],Ptemp[1,0]),np.append(Ang,Ang[0]+2.*np.pi)],
        # w=ww, s=ss, k=kk, nest=nest)
        tckp, uu = scpinterp.splprep(
            [
                np.append(Ptemp[0, :], Ptemp[0, 0]),
                np.append(Ptemp[1, :], Ptemp[1, 0]),
            ],
            u=np.append(Ang, Ang[0] + 2.0 * np.pi),
            w=ww,
            s=ss,
            k=kk,
            nest=nest,
            full_output=0,
        )
        xnew, ynew = scpinterp.splev(np.linspace(-np.pi, np.pi, NP), tckp)
        Poly = np.array([xnew + BaryS[0], ynew + BaryS[1]])
        Poly = np.concatenate((Poly, Poly[:, 0:1]), axis=1)
    if Plot:
        import matplotlib.pyplot as plt

        f = plt.figure(facecolor="w", figsize=(8, 10))
        ax = f.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.plot(Poly[0, :], Poly[1, :], "-k", Poly[0, :], Poly[1, :], "-r")
        ax.set_aspect(aspect="equal", adjustable="datalim"), ax.set_xlabel(
            r"R (m)"
        ), ax.set_ylabel(r"Z (m)")
        f.canvas.draw()
    return Poly


# ==============================================================================
# = Ves sampling functions
# ==============================================================================


def _Ves_get_sample_checkinputs(
    res=None,
    domain=None,
    resMode=None,
    ind=None,
    which='volume'
):
    """ Check inputs for all sampling routines """

    # res
    dres = {'edge': 1, 'cross': 2, 'surface': 2, 'volume': 3}
    if res is None:
        res = _SAMPLE_RES[which]
    ltypes = [int, float, np.int_, np.float64]
    c0 = (type(res) in ltypes
          or (hasattr(res, "__iter__")
              and len(res) == dres[which]
              and all([type(ds) in ltypes for ds in res])))
    if not c0:
        msg = ("Arg res must be either:\n"
               + "\t- float: unique resolution for all directions\n"
               + "\t- iterable of {} floats\n".format(dres[which])
               + "  You provided:\n{}".format(res))
        raise Exception(msg)
    if type(res) in ltypes:
        if which != 'edge':
            res = [float(res) for ii in range(dres[which])]
    else:
        if which == 'edge':
            msg = ("For edge, res cannot be an iterable!\n"
                   + "\t- res: {}".format(res))
            raise Exception(msg)
        res = [float(res[ii]) for ii in range(dres[which])]

    # domain (i.e.: sub-domain to be sampled, defined by its limits)
    ddomain = {'edge': 2, 'cross': 2, 'surface': 3, 'volume': 3}
    if domain is None:
        domain = [None for ii in range(ddomain[which])]
    c0 = (hasattr(domain, "__iter__")
          and len(domain) == ddomain[which]
          and all([dd is None
                   or (hasattr(dd, "__iter__")
                       and len(dd) == 2
                       and all([ss is None
                                or type(ss) in ltypes
                                for ss in dd]))
                   for dd in domain]))
    if not c0:
        msg = ("Arg domain must be a len()={} iterable".format(ddomain[which])
               + " where each element can be:\n"
               + "\t- an iterable 2 floats: [lower, upper] bounds\n"
               + "\t- None: no bounds\n"
               + "  You provided:\n{}".format(domain))
        raise Exception(msg)
    for ii in range(len(domain)):
        if domain[ii] is not None:
            domain[ii] = [float(domain[ii][0])
                          if domain[ii][0] is not None else None,
                          float(domain[ii][1])
                          if domain[ii][1] is not None else None]

    # resMode
    if resMode is None:
        resMode = _SAMPLE_RESMODE[which]
    c0 = isinstance(resMode, str) and resMode.lower() in ["abs", "rel"]
    if not c0:
        msg = ("Arg resMode must be in ['abs','rel']!\n"
               + "  You provided:\n{}".format(resMode))
        raise Exception(msg)
    resMode = resMode.lower()

    # ind (indices of points to be recovered)
    c0 = (ind is None
          or (isinstance(ind, np.ndarray)
              and ind.ndim == 1
              and 'int' in ind.dtype.name
              and np.all(ind >= 0))
          or (which == 'surface'
              and isinstance(ind, list)
              and all([isinstance(indi, np.ndarray)
                       and indi.ndim == 1
                       and 'int' in indi.dtype.name
                       and np.all(indi >= 0) for indi in ind])))
    if not c0:
        msg = ("Arg ind must be either:\n"
               + "\t- None: domain is used instead\n"
               + "\t- 1d np.ndarray of positive int: indices\n")
        if which == 'surface':
            msg += "\t- list of 1d np.ndarray of positive indices\n"
        msg += "  You provided:\n{}".format(ind)
        raise Exception(msg)

    if isinstance(ind, np.ndarray):
        ind = ind.astype(np.int64)
    elif isinstance(ind, list):
        ind = [ii.astype(np.int64) for ii in ind]

    return res, domain, resMode, ind


def _Ves_get_sampleEdge(
    VPoly,
    res=None,
    domain=None,
    resMode=None,
    offsetIn=0.0,
    VIn=None,
    margin=1.0e-9
):

    # -------------
    #  Check inputs

    # standard
    res, domain, resMode, ind = _Ves_get_sample_checkinputs(
        res=res,
        domain=domain,
        resMode=resMode,
        ind=None,
        which='edge',
    )

    # specific
    ltypes = [int, float, np.int_, np.float64]
    assert type(offsetIn) in ltypes

    # -------------
    #  Compute

    Pts, reseff, ind, N, Rref, VPolybis = _GG.discretize_vpoly(
        VPoly,
        float(res),
        mode=resMode,
        D1=domain[0],
        D2=domain[1],
        margin=margin,
        DIn=float(offsetIn),
        VIn=VIn,
    )
    return Pts, reseff, ind


def _Ves_get_sampleCross(
    VPoly,
    Min1,
    Max1,
    Min2,
    Max2,
    res=None,
    domain=None,
    resMode=None,
    ind=None,
    margin=1.0e-9,
    mode="flat",
):

    # -------------
    #  Check inputs

    # standard
    res, domain, resMode, ind = _Ves_get_sample_checkinputs(
        res=res,
        domain=domain,
        resMode=resMode,
        ind=ind,
        which='cross',
    )

    # specific
    assert mode in ["flat", "imshow"]

    # -------------
    #  Compute

    MinMax1 = np.array([Min1, Max1])
    MinMax2 = np.array([Min2, Max2])
    if ind is None:
        if mode == "flat":
            Pts, dS, ind, d1r, d2r = _GG.discretize_segment2d(
                MinMax1,
                MinMax2,
                res[0],
                res[1],
                D1=domain[0],
                D2=domain[1],
                mode=resMode,
                VPoly=VPoly,
                margin=margin,
            )
            out = (Pts, dS, ind, (d1r, d2r))
        else:
            x1, d1r, ind1, N1 = _GG._Ves_mesh_dlfromL_cython(
                MinMax1, res[0], domain[0], Lim=True,
                dLMode=resMode, margin=margin
            )
            x2, d2r, ind2, N2 = _GG._Ves_mesh_dlfromL_cython(
                MinMax2, res[1], domain[1], Lim=True,
                dLMode=resMode, margin=margin
            )
            xx1, xx2 = np.meshgrid(x1, x2)
            pts = np.squeeze([xx1, xx2])
            extent = (
                x1[0] - d1r / 2.0,
                x1[-1] + d1r / 2.0,
                x2[0] - d2r / 2.0,
                x2[-1] + d2r / 2.0,
            )
            out = (pts, x1, x2, extent)
    else:
        assert mode == "flat"
        c0 = type(ind) is np.ndarray and ind.ndim == 1
        c0 = c0 and ind.dtype in ["int32", "int64"] and np.all(ind >= 0)
        assert c0, "Arg ind must be a np.ndarray of int !"
        Pts, dS, d1r, d2r = _GG._Ves_meshCross_FromInd(
            MinMax1,
            MinMax2,
            res[0],
            res[1],
            ind,
            dSMode=resMode,
            margin=margin,
        )
        out = (Pts, dS, ind, (d1r, d2r))
    return out


def _Ves_get_sampleS(
    VPoly,
    res=None,
    domain=None,
    resMode="abs",
    ind=None,
    offsetIn=0.0,
    VIn=None,
    VType="Tor",
    VLim=None,
    nVLim=None,
    returnas="(X,Y,Z)",
    margin=1.0e-9,
    Multi=False,
    Ind=None,
):
    """ Sample the surface """

    # -------------
    #  Check inputs

    # standard
    res, domain, resMode, ind = _Ves_get_sample_checkinputs(
        res=res,
        domain=domain,
        resMode=resMode,
        ind=ind,
        which='surface',
    )

    # nVLim and VLim
    if not (type(nVLim) in [int, np.int_] and nVLim >= 0):
        msg = ("Arg nVLim must be a positive int\\n"
               + "  You provided:\n{} ({})".format(nVLim, type(nVLim)))
        raise Exception(msg)
    VLim = None if (VLim is None or nVLim == 0) else np.array(VLim)

    if not isinstance(Multi, bool):
        msg = ("Arg Multi must be a bool!\n"
               + "  You provided:\n{}".format(Multi))
        raise Exception(msg)

    # Check if Multi
    if nVLim > 1:
        assert VLim is not None, "For multiple Struct, Lim cannot be None !"
        assert all([hasattr(ll, "__iter__") and len(ll) == 2 for ll in VLim])
        if Ind is None:
            Ind = np.arange(0, nVLim)
        else:
            Ind = [Ind] if not hasattr(Ind, "__iter__") else Ind
            Ind = np.asarray(Ind).astype(int)
        if ind is not None:
            if isinstance(ind, np.ndarray):
                ind = [ind for ii in range(len(Ind))]
            elif not (isinstance(ind, list) and len(ind) == len(Ind)):
                msg = ("Arg ind must be a list of same len() as Ind!\n"
                       + "\t- provided: {}".format(ind))
                raise Exception(msg)

    else:
        VLim = [None] if VLim is None else [VLim.ravel()]
        if ind is not None:
            if not isinstance(ind, np.ndarray):
                msg = ("ind must be a np.ndarray if nVLim == 1\n"
                       + "\t- provided: {}".format(ind))
                raise Exception(msg)
        Ind = [0]

    if ind is None:
        pts, dS, ind, reseff = (
            [0 for ii in Ind],
            [0 for ii in Ind],
            [0 for ii in Ind],
            [[0, 0] for ii in Ind],
        )
        if VType.lower() == "tor":
            for ii in range(0, len(Ind)):
                if VLim[Ind[ii]] is None:
                    (pts[ii],
                     dS[ii],
                     ind[ii],
                     NL,
                     reseff[ii][0],
                     Rref,
                     reseff[ii][1],
                     nRPhi0,
                     VPbis) = _GG._Ves_Smesh_Tor_SubFromD_cython(
                         res[0],
                         res[1],
                         VPoly,
                         DR=domain[0],
                         DZ=domain[1],
                         DPhi=domain[2],
                         DIn=offsetIn,
                         VIn=VIn,
                         PhiMinMax=None,
                         Out=returnas,
                         margin=margin,
                     )
                else:
                    (pts[ii],
                     dS[ii],
                     ind[ii],
                     NL,
                     reseff[ii][0],
                     Rref,
                     dR0r,
                     dZ0r,
                     reseff[ii][1],
                     VPbis) = _GG._Ves_Smesh_TorStruct_SubFromD_cython(
                         VLim[Ind[ii]],
                         res[0],
                         res[1],
                         VPoly,
                         DR=domain[0],
                         DZ=domain[1],
                         DPhi=domain[2],
                         DIn=offsetIn,
                         VIn=VIn,
                         Out=returnas,
                         margin=margin,
                     )
                    reseff[ii] += [dR0r, dZ0r]
        else:
            for ii in range(0, len(Ind)):
                (pts[ii],
                 dS[ii],
                 ind[ii],
                 NL,
                 reseff[ii][0],
                 Rref,
                 reseff[ii][1],
                 dY0r,
                 dZ0r,
                 VPbis) = _GG._Ves_Smesh_Lin_SubFromD_cython(
                     VLim[Ind[ii]],
                     res[0],
                     res[1],
                     VPoly,
                     DX=domain[0],
                     DY=domain[1],
                     DZ=domain[2],
                     DIn=offsetIn,
                     VIn=VIn,
                     margin=margin,
                 )
                reseff[ii] += [dY0r, dZ0r]
    else:
        ind = ind if Multi else [ind]
        pts, dS, reseff = (
            [np.ones((3, 0)) for ii in Ind],
            [0 for ii in Ind],
            [[0, 0] for ii in Ind],
        )
        if VType.lower() == "tor":
            for ii in range(0, len(Ind)):
                if ind[Ind[ii]].size > 0:
                    if VLim[Ind[ii]] is None:
                        out_loc = _GG._Ves_Smesh_Tor_SubFromInd_cython(
                            res[0],
                            res[1],
                            VPoly,
                            ind[Ind[ii]],
                            DIn=offsetIn,
                            VIn=VIn,
                            PhiMinMax=None,
                            Out=returnas,
                            margin=margin,
                        )
                        pts[ii], dS[ii], NL, reseff[ii][0], Rref = out_loc[:5]
                        reseff[ii][1], nRPhi0, VPbis = out_loc[5:]
                    else:
                        out_loc = _GG._Ves_Smesh_TorStruct_SubFromInd_cython(
                            VLim[Ind[ii]],
                            res[0],
                            res[1],
                            VPoly,
                            ind[Ind[ii]],
                            DIn=offsetIn,
                            VIn=VIn,
                            Out=returnas,
                            margin=margin,
                        )
                        pts[ii], dS[ii], NL, reseff[ii][0], Rref = out_loc[:5]
                        dR0r, dZ0r, reseff[ii][1], VPbis = out_loc[5:]
                        reseff[ii] += [dR0r, dZ0r]
        else:
            for ii in range(0, len(Ind)):
                if ind[Ind[ii]].size > 0:
                    out_loc = _GG._Ves_Smesh_Lin_SubFromInd_cython(
                        VLim[Ind[ii]],
                        res[0],
                        res[1],
                        VPoly,
                        ind[Ind[ii]],
                        DIn=offsetIn,
                        VIn=VIn,
                        margin=margin,
                    )
                    pts[ii], dS[ii], NL, reseff[ii][0], Rref = out_loc[:5]
                    reseff[ii][1], dY0r, dZ0r, VPbis = out_loc[5:]
                    reseff[ii] += [dY0r, dZ0r]

    if len(VLim) == 1:
        pts, dS, ind, reseff = pts[0], dS[0], ind[0], reseff[0]
    return pts, dS, ind, reseff


def _Ves_get_sampleV(
    VPoly,
    Min1,
    Max1,
    Min2,
    Max2,
    res=None,
    domain=None,
    resMode=None,
    ind=None,
    VType="Tor",
    VLim=None,
    returnas="(X,Y,Z)",
    margin=1.0e-9,
    algo="new",
    num_threads=48,
):
    """ Sample the volume """

    # -------------
    #  Check inputs
    res, domain, resMode, ind = _Ves_get_sample_checkinputs(
        res=res,
        domain=domain,
        resMode=resMode,
        ind=ind,
        which='volume',
    )

    # ------------
    # Computation

    MinMax1 = np.array([Min1, Max1])
    MinMax2 = np.array([Min2, Max2])
    VLim = None if VType.lower() == "tor" else np.array(VLim).ravel()
    reseff = [None, None, None]
    if ind is None:
        if VType.lower() == "tor":
            if algo.lower() == "new":
                (pts, dV, ind,
                 reseff[0],
                 reseff[1],
                 reseff[2],
                 sz_r, sz_z,
                 ) = _GG._Ves_Vmesh_Tor_SubFromD_cython(
                    res[0],
                    res[1],
                    res[2],
                    MinMax1,
                    MinMax2,
                    DR=domain[0],
                    DZ=domain[1],
                    DPhi=domain[2],
                    limit_vpoly=VPoly,
                    out_format=returnas,
                    margin=margin,
                    num_threads=num_threads,
                )
            else:
                (pts, dV, ind,
                 reseff[0],
                 reseff[1],
                 reseff[2]) = _GG._Ves_Vmesh_Tor_SubFromD_cython_old(
                    res[0],
                    res[1],
                    res[2],
                    MinMax1,
                    MinMax2,
                    DR=domain[0],
                    DZ=domain[1],
                    DPhi=domain[2],
                    VPoly=VPoly,
                    Out=returnas,
                    margin=margin,
                )

        else:
            (pts, dV, ind,
             reseff[0],
             reseff[1],
             reseff[2]) = _GG._Ves_Vmesh_Lin_SubFromD_cython(
                res[0],
                res[1],
                res[2],
                VLim,
                MinMax1,
                MinMax2,
                DX=domain[0],
                DY=domain[1],
                DZ=domain[2],
                limit_vpoly=VPoly,
                margin=margin,
            )
    else:
        if VType.lower() == "tor":
            if algo.lower() == "new":
                (pts, dV,
                 reseff[0],
                 reseff[1],
                 reseff[2]) = _GG._Ves_Vmesh_Tor_SubFromInd_cython(
                    res[0],
                    res[1],
                    res[2],
                    MinMax1,
                    MinMax2,
                    ind,
                    Out=returnas,
                    margin=margin,
                    num_threads=num_threads,
                )
            else:
                (pts, dV,
                 reseff[0],
                 reseff[1],
                 reseff[2]) = _GG._Ves_Vmesh_Tor_SubFromInd_cython_old(
                    res[0],
                    res[1],
                    res[2],
                    MinMax1,
                    MinMax2,
                    ind,
                    Out=returnas,
                    margin=margin,
                )
        else:
            (pts, dV,
             reseff[0],
             reseff[1],
             reseff[2]) = _GG._Ves_Vmesh_Lin_SubFromInd_cython(
                 res[0],
                 res[1],
                 res[2],
                 VLim,
                 MinMax1,
                 MinMax2,
                 ind,
                 margin=margin
             )
    return pts, dV, ind, reseff


# ==============================================================================
# =  phi / theta projections for magfieldlines
# ==============================================================================


def _Struct_get_phithetaproj(ax=None, poly_closed=None, lim=None, noccur=0):

    # phi = toroidal angle
    if noccur == 0:
        Dphi = np.array([[-np.pi, np.pi]])
        nphi = np.r_[1]
    else:
        assert lim.ndim == 2, str(lim)
        nphi = np.ones((noccur,), dtype=np.int64)
        ind = (lim[:, 0] > lim[:, 1]).nonzero()[0]
        Dphi = np.concatenate((lim, np.full((noccur, 2), np.nan)), axis=1)
        if ind.size > 0:
            for ii in ind:
                Dphi[ii, :] = [lim[ii, 0], np.pi, -np.pi, lim[ii, 1]]
                nphi[ii] = 2

    # theta = poloidal angle
    Dtheta = np.arctan2(poly_closed[1, :] - ax[1], poly_closed[0, :] - ax[0])
    Dtheta = np.r_[np.min(Dtheta), np.max(Dtheta)]
    if Dtheta[0] > Dtheta[1]:
        ntheta = 2
        Dtheta = [Dtheta[0], np.pi, -np.pi, Dtheta[1]]
    else:
        ntheta = 1

    return nphi, Dphi, ntheta, Dtheta


def _get_phithetaproj_dist(
    poly_closed,
    ax,
    Dtheta,
    nDtheta,
    Dphi,
    nDphi,
    theta,
    phi,
    ntheta,
    nphi,
    noccur,
):

    if nDtheta == 1:
        ind = (theta >= Dtheta[0]) & (theta <= Dtheta[1])
    else:
        ind = (theta >= Dtheta[0]) | (theta <= Dtheta[1])

    disttheta = np.full((theta.size,), np.nan)

    # phi within Dphi
    if noccur > 0:
        indphi = np.zeros((nphi,), dtype=bool)
        for ii in range(0, noccur):
            for jj in range(0, nDphi[ii]):
                indphi |= (phi >= Dphi[ii, jj]) & (phi <= Dphi[ii, jj + 1])
        if not np.any(indphi):
            return disttheta, indphi
    else:
        indphi = np.ones((nphi,), dtype=bool)

    # No theta within Dtheta
    if not np.any(ind):
        return disttheta, indphi

    # Check for non-parallel AB / u pairs
    u = np.array([np.cos(theta), np.sin(theta)])
    AB = np.diff(poly_closed, axis=1)
    detABu = AB[0, :, None] * u[1, None, :] - AB[1, :, None] * u[0, None, :]
    inddet = ind[None, :] & (np.abs(detABu) > 1.0e-9)
    if not np.any(inddet):
        return disttheta, indphi

    nseg = poly_closed.shape[1] - 1
    k = np.full((nseg, ntheta), np.nan)

    OA = poly_closed[:, :-1] - ax[:, None]
    detOAu = (OA[0, :, None] * u[1, None, :] - OA[1, :, None] * u[0, None, :])[
        inddet
    ]
    ss = -detOAu / detABu[inddet]
    inds = (ss >= 0.0) & (ss < 1.0)
    inddet[inddet] = inds

    if not np.any(inds):
        return disttheta, indphi

    scaOAu = (OA[0, :, None] * u[0, None, :] + OA[1, :, None] * u[1, None, :])[
        inddet
    ]
    scaABu = (AB[0, :, None] * u[0, None, :] + AB[1, :, None] * u[1, None, :])[
        inddet
    ]
    k[inddet] = scaOAu + ss[inds] * scaABu
    indk = k[inddet] > 0.0
    inddet[inddet] = indk

    if not np.any(indk):
        return disttheta, indphi

    k[~inddet] = np.nan
    indok = np.any(inddet, axis=0)
    disttheta[indok] = np.nanmin(k[:, indok], axis=0)

    return disttheta, indphi


# ==============================================================================
# =  LOS functions
# ==============================================================================
def LOS_PRMin(Ds, us, kOut=None, Eps=1.0e-12, squeeze=True, Test=True):
    """  Compute the point on the LOS where the major radius is minimum """
    if Test:
        assert Ds.ndim in [1, 2, 3] and 3 in Ds.shape and Ds.shape == us.shape
    if kOut is not None:
        kOut = np.atleast_1d(kOut)
        assert kOut.size == Ds.size / 3

    if Ds.ndim == 1:
        Ds, us = Ds[:, None, None], us[:, None, None]
    elif Ds.ndim == 2:
        Ds, us = Ds[:, :, None], us[:, :, None]
    if kOut is not None:
        if kOut.ndim == 1:
            kOut = kOut[:, None]
    _, nlos, nref = Ds.shape

    kRMin = np.full((nlos, nref), np.nan)
    uparN = np.sqrt(us[0, :, :] ** 2 + us[1, :, :] ** 2)

    # Case with u vertical
    ind = uparN > Eps
    kRMin[~ind] = 0.0

    # Else
    kRMin[ind] = (
        -(us[0, ind] * Ds[0, ind] + us[1, ind] * Ds[1, ind]) / uparN[ind] ** 2
    )

    # Check
    kRMin[kRMin <= 0.0] = 0.0
    if kOut is not None:
        kRMin[kRMin > kOut] = kOut[kRMin > kOut]

    # squeeze
    if squeeze:
        if nref == 1 and nlos == 11:
            kRMin = kRMin[0, 0]
        elif nref == 1:
            kRMin = kRMin[:, 0]
        elif nlos == 1:
            kRMin = kRMin[0, :]
    return kRMin


def LOS_CrossProj(
    VType,
    Ds,
    us,
    kOuts,
    proj="All",
    multi=False,
    num_threads=16,
    return_pts=False,
    Test=True,
):
    """ Compute the parameters to plot the poloidal projection of the LOS  """
    assert type(VType) is str and VType.lower() in ["tor", "lin"]
    dproj = {
        "cross": ("R", "Z"),
        "hor": ("x,y"),
        "all": ("R", "Z", "x", "y"),
        "3d": ("x", "y", "z"),
    }
    assert type(proj) in [str, tuple]
    if type(proj) is tuple:
        assert all([type(pp) is str for pp in proj])
        lcoords = proj
    else:
        proj = proj.lower()
        assert proj in dproj.keys()
        lcoords = dproj[proj]
    if return_pts:
        assert proj in ["cross", "hor", "3d"]

    lc = [Ds.ndim == 3, Ds.shape == us.shape]
    if not all(lc):
        msg = "Ds and us must have the same shape and dim in [2,3]:\n"
        msg += "    - provided Ds.shape: %s\n" % str(Ds.shape)
        msg += "    - provided us.shape: %s" % str(us.shape)
        raise Exception(msg)
    lc = [kOuts.size == Ds.size / 3, kOuts.shape == Ds.shape[1:]]
    if not all(lc):
        msg = "kOuts must have the same shape and ndim = Ds.ndim-1:\n"
        msg += "    - Ds.shape    : %s\n" % str(Ds.shape)
        msg += "    - kOutss.shape: %s" % str(kOuts.shape)
        raise Exception(msg)

    # Prepare inputs
    _, nlos, nseg = Ds.shape

    # Detailed sampling for 'tor' and ('cross' or 'all')
    R, Z = None, None
    if "R" in lcoords or "Z" in lcoords:
        angcross = np.arccos(
            np.sqrt(us[0, ...] ** 2 + us[1, ...] ** 2)
            / np.sqrt(np.sum(us ** 2, axis=0))
        )
        resnk = np.ceil(25.0 * (1 - (angcross / (np.pi / 4) - 1) ** 2) + 5)
        resnk = 1.0 / resnk.ravel()

        # Use optimized get sample
        DL = np.vstack((np.zeros((nlos * nseg,), dtype=float), kOuts.ravel()))

        k, reseff, lind = _GG.LOS_get_sample(
            nlos * nseg,
            resnk,
            DL,
            dmethod="rel",
            method="simps",
            num_threads=num_threads,
            Test=Test,
        )

        assert lind.size == nseg * nlos - 1
        ind = lind[nseg - 1 :: nseg]  # noqa
        nbrep = np.r_[lind[0], np.diff(lind), k.size - lind[-1]]
        pts = np.repeat(Ds.reshape((3, nlos * nseg)), nbrep, axis=1) + k[
            None, :
        ] * np.repeat(us.reshape((3, nlos * nseg)), nbrep, axis=1)

        if return_pts:
            pts = np.array([np.hypot(pts[0, :], pts[1, :]), pts[2, :]])
            if multi:
                pts = np.split(pts, ind, axis=1)
            else:
                pts = np.insert(pts, ind, np.nan, axis=1)
        else:
            if multi:
                if "R" in lcoords:
                    R = np.split(np.hypot(pts[0, :], pts[1, :]), ind)
                if "Z" in lcoords:
                    Z = np.split(pts[2, :], ind)
            else:
                if "R" in lcoords:
                    R = np.insert(np.hypot(pts[0, :], pts[1, :]), ind, np.nan)
                if "Z" in lcoords:
                    Z = np.insert(pts[2, :], ind, np.nan)

    # Normal sampling => pts
    # unnecessary only if 'tor' and 'cross'
    x, y, z = None, None, None
    if "x" in lcoords or "y" in lcoords or "z" in lcoords:
        pts = np.concatenate(
            (Ds, Ds[:, :, -1:] + kOuts[None, :, -1:] * us[:, :, -1:]), axis=-1
        )

        if multi:
            ind = np.arange(1, nlos) * (nseg + 1)
            pts = pts.reshape((3, nlos * (nseg + 1)))
        else:
            nancoords = np.full((3, nlos, 1), np.nan)
            pts = np.concatenate((pts, nancoords), axis=-1)
            pts = pts.reshape((3, nlos * (nseg + 2)))

        if return_pts:
            assert proj in ["hor", "3d"]
            if multi:
                if proj == "hor":
                    pts = np.split(pts[:2, :], ind, axis=1)
                else:
                    pts = np.split(pts, ind, axis=1)
            elif proj == "hor":
                pts = pts[:2, :]

        else:
            if multi:
                if "x" in lcoords:
                    x = np.split(pts[0, :], ind)
                if "y" in lcoords:
                    y = np.split(pts[1, :], ind)
                if "z" in lcoords:
                    z = np.split(pts[2, :], ind)
            else:
                if "x" in lcoords:
                    x = pts[0, :]
                if "y" in lcoords:
                    y = pts[1, :]
                if "z" in lcoords:
                    z = pts[2, :]

    if return_pts:
        return pts
    else:
        return R, Z, x, y, z


# ==============================================================================
# =  Meshing & signal
# ==============================================================================
def LOS_get_sample(D, u, dL, DL=None, dLMode="abs", method="sum", Test=True):
    """ Return the sampled line, with the specified method

    'linspace': return the N+1 edges, including the first and last point
    'sum' : return the N middle of the segments
    'simps': return the N+1 egdes, where N has to be even
             (scipy.simpson requires an even number of intervals)
    'romb' : return the N+1 edges, where N+1 = 2**k+1
             (fed to scipy.romb for integration)
    """
    if Test:
        assert all(
            [type(dd) is np.ndarray and dd.shape == (3,) for dd in [D, u]]
        )
        assert not hasattr(dL, "__iter__")
        assert DL is None or all(
            [
                hasattr(DL, "__iter__"),
                len(DL) == 2,
                all([not hasattr(dd, "__iter__") for dd in DL]),
            ]
        )
        assert dLMode in ["abs", "rel"]
        assert type(method) is str and method in [
            "linspace",
            "sum",
            "simps",
            "romb",
        ]
    # Compute the min number of intervals to satisfy the specified resolution
    N = (
        int(np.ceil((DL[1] - DL[0]) / dL))
        if dLMode == "abs"
        else int(np.ceil(1.0 / dL))
    )
    # Modify N according to the desired method
    if method == "simps":
        N = N if N % 2 == 0 else N + 1
    elif method == "romb":
        N = 2 ** int(np.ceil(np.log(N) / np.log(2.0)))

    # Derive k and dLr
    if method == "sum":
        dLr = (DL[1] - DL[0]) / N
        k = DL[0] + (0.5 + np.arange(0, N)) * dLr
    else:
        k, dLr = np.linspace(
            DL[0], DL[1], N + 1, endpoint=True, retstep=True, dtype=float
        )

    Pts = D[:, np.newaxis] + k[np.newaxis, :] * u[:, np.newaxis]
    return Pts, k, dLr


def LOS_calc_signal(
    ff, D, u, dL, DL=None, dLMode="abs", method="romb", Test=True
):
    assert hasattr(ff, "__call__"), (
        "Arg ff must be a callable (function) taking at least 1 positional ",
        "Pts (a (3,N) np.ndarray of cartesian (X,Y,Z) coordinates) !",
    )
    assert not method == "linspace"
    Pts, k, dLr = LOS_get_sample(
        D, u, dL, DL=DL, dLMode=dLMode, method=method, Test=Test
    )
    out = insp(ff)
    N = np.sum(
        [
            (
                pp.kind == pp.POSITIONAL_OR_KEYWORD
                and pp.default is pp.empty
            )
            for pp in out.parameters.values()
        ]
    )

    if N == 1:
        Vals = ff(Pts)
    elif N == 2:
        Vals = ff(Pts, np.tile(-u, (Pts.shape[1], 1)).T)
    else:
        raise ValueError(
            "The function (ff) assessing the emissivity locally "
            + "must take a single positional argument: Pts a (3,N)"
            + " np.ndarray of (X,Y,Z) cartesian coordinates !"
        )

    Vals[np.isnan(Vals)] = 0.0
    if method == "sum":
        Int = np.sum(Vals) * dLr
    elif method == "simps":
        Int = scpintg.simpson(Vals, x=None, dx=dLr)
    elif method == "romb":
        Int = scpintg.romb(Vals, dx=dLr, show=False)
    return Int


# -*- coding: utf-8 -*-


import copy
import warnings


import numpy as np
import scipy.linalg as scplin
import scipy.optimize as scpop
import scipy.sparse as scpsp
import datastock as ds


dfail = {}
try:
    import sksparse as sksp
except Exception as err:
    sksp = False
    dfail['sksparse'] = "For cholesk factorizations"

try:
    import scikits.umfpack as skumf
except Exception as err:
    skumf = False
    dfail['umfpack'] = "For faster sparse matrices"

if len(dfail) > 0:
    lstr = [f"\t- {k0}: {v0}" for k0, v0 in dfail.items()]
    msg = (
        "Consider installing the following for faster inversions:\n"
        + "\n".join(lstr)
    )
    warnings.warn(msg)


# optional
try:
    from .. import tomotok2tofu
except Exception as err:
    tomotok2tofu = False


_DALGO0 = {
    'algo0': {
        'source': 'tofu',
        'family': 'Phillips-Tikhonov',
        'reg_operator': 'any linear',
        'reg_param': 'augTikho',
        'decomposition': '',
        'positive': False,
        'sparse': True,
        'isotropic': True,
        'func': 'inv_linear_augTikho_sparse',
    },
    'algo1': {
        'source': 'tofu',
        'family': 'Phillips-Tikhonov',
        'reg_operator': 'any linear',
        'reg_param': 'augTikho',
        'decomposition': '',
        'positive': False,
        'sparse': False,
        'isotropic': True,
        'func': 'inv_linear_augTikho_dense',
    },
    'algo2': {
        'source': 'tofu',
        'family': 'Phillips-Tikhonov',
        'reg_operator': 'any linear',
        'reg_param': 'augTikho',
        'decomposition': 'cholesky',
        'positive': False,
        'sparse': False,
        'isotropic': True,
        'func': 'inv_linear_augTikho_chol_dense',
    },
    'algo3': {
        'source': 'tofu',
        'family': 'Phillips-Tikhonov',
        'reg_operator': 'any linear',
        'reg_param': 'augTikho',
        'decomposition': 'cholesky',
        'positive': False,
        'sparse': True,
        'isotropic': True,
        'func': 'inv_linear_augTikho_chol_sparse',
    },
    'algo4': {
        'source': 'tofu',
        'family': 'Phillips-Tikhonov',
        'reg_operator': 'any linear',
        'reg_param': 'augTikho',
        'decomposition': '',
        'positive': True,
        'sparse': False,
        'isotropic': True,
        'func': 'inv_linear_augTikho_pos_dense',
    },
    'algo5': {
        'source': 'tofu',
        'family': 'Phillips-Tikhonov',
        'reg_operator': 'any linear',
        'reg_param': 'DisPrinc',
        'decomposition': '',
        'positive': False,
        'sparse': True,
        'isotropic': True,
        'func': 'inv_linear_DisPrinc_sparse',
    },
    'algo6': {
        'source': 'tofu',
        'family': 'Non-regularized',
        'reg_operator': None,
        'reg_param': None,
        'decomposition': '',
        'positive': False,
        'sparse': True,
        'isotropic': True,
        'func': 'inv_linear_leastsquares_bounds',
    },
    'algo7': {
        'source': 'tofu',
        'family': 'Non-regularized',
        'reg_operator': None,
        'reg_param': None,
        'decomposition': '',
        'positive': True,
        'sparse': True,
        'isotropic': True,
        'func': 'inv_linear_leastsquares_bounds',
    },
}


# #############################################################################
# #############################################################################
#                           info
# #############################################################################


def get_available_inversions_algo(
    # for filtering
    source=None,
    family=None,
    reg_operator=None,
    reg_param=None,
    decomposition=None,
    positive=None,
    sparse=None,
    isotropic=None,
    dalgo=None,
    # parameters
    returnas=None,
    verb=None,
):

    # --------------
    # check inputs

    returnas = ds._generic_check._check_var(
        returnas,
        'returnas',
        default=False,
        allowed=[False, dict, list, str]
    )

    verb = ds._generic_check._check_var(
        verb,
        'verb',
        default=returnas is False,
        types=bool,
    )

    # ------------------------------------
    # filter according to criteria, if any

    dalgo = match_algo(
        # filtering
        source=source,
        family=family,
        reg_operator=reg_operator,
        reg_param=reg_param,
        decomposition=decomposition,
        positive=positive,
        sparse=sparse,
        isotropic=isotropic,
    )

    # complemet with name
    for k0 in dalgo.keys():
        dalgo[k0]['name'] = k0

    # ------------
    # print or str

    if verb is True or returnas is str:

        head = ['key'] + [
            'source', 'family', 'reg_operator', 'reg_param', 'decomposition',
            'positive', 'sparse',
        ]
        sep = ['-'*len(kk) for kk in head]
        lstr = [head, sep] + [
            [k0] + [v0[k1] for k1 in head[1:]]
            for k0, v0 in dalgo.items()
        ]

        nmax = np.max(np.char.str_len(np.char.array(lstr)), axis=0)
        lstr = [
            '  '.join([str(ss).ljust(nmax[ii]) for ii, ss in enumerate(line)])
            for line in lstr
        ]
        msg = "\n".join(lstr)

        if verb:
            print(msg)

    # -------
    # return

    if returnas is dict:
        return dalgo
    elif returnas is list:
        return sorted(dalgo.keys())
    elif returnas is str:
        return msg


def match_algo(
    source=None,
    family=None,
    reg_operator=None,
    reg_param=None,
    decomposition=None,
    positive=None,
    sparse=None,
    isotropic=None,
):

    # ------------
    # check inputs

    dargs = {
        k0: v0 for k0, v0 in locals().items()
        if v0 is not None
    }

    # --------------
    # Get tofu algo

    dalgo = _DALGO0

    # -----------------
    # Get tomotok algo

    if tomotok2tofu is not False:
        dalgo.update(tomotok2tofu.get_dalgo())

    # ------------
    # find matches

    if len(dargs) > 0:
        lmatch = [
            k0 for k0, v0 in dalgo.items()
            if all([v0[k1] == v1 for k1, v1 in dargs.items()])
        ]
        if len(lmatch) == 0:
            lstr = [f'\t- {k0}: {v0}' for k0, v0 in dargs.items()]
            msg = (
                "No / several algorithms matching the following criteria:\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)
        dalgo = {k0: v0 for k0, v0 in dalgo.items() if k0 in lmatch}

    return copy.deepcopy(dalgo)


# ################################################################
# ################################################################
#                      Basic routines - augmented tikhonov
# ################################################################


def inv_linear_augTikho_dense(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    nbs=None,
    mu0=None,
    conv_crit=None,
    a0bis=None,
    b0=None,
    a1bis=None,
    b1=None,
    d=None,
    conv_reg=True,
    verb=None,
    verb2head=None,
    maxiter_outer=None,
    **kwdargs,
):
    """
    Linear algorithm for Phillips-Tikhonov regularisation
    Called "Augmented Tikhonov", dense matrix version
    """

    conv = 0.           # convergence variable
    niter = 0           # number of iterations
    mu1 = 0.            # regularisation param

    # verb
    if verb >= 2:
        chi2n = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {chi2n:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*chi2n + mu0*reg:.3e}",
            end='\n',
        )

    # loop
    # Continue until convergence criterion, and at least 2 iterations
    while niter <= 2 or (conv > conv_crit and niter < maxiter_outer):

        # call solver
        sol = scplin.solve(
            TTn + mu0*R, Tyn,
            assume_a='pos',      # 'pos' faster than 'sym'
            overwrite_a=True,    # no significant gain
            overwrite_b=False,   # True is faster, but a copy of Tyn is needed
            check_finite=False,  # small speed gain compared to True
            transposed=False,
        )  # 3

        # call augmented Tikhonov update of mu
        mu1, conv, res2, reg, tau, lamb = _augTikho_update(
            Tn, sol, yn, R,
            a0bis, b0, a1bis, b1,
            d, mu0, conv_reg, nbs, sol0,
            # verb
            verb=verb,
            nchan=nchan,
            niter=niter,
        )

        # update sol0, mu0 for next iteration
        sol0[:] = sol[:]
        mu0 = mu1
        niter += 1
    return sol, mu1, res2/nchan, reg, niter, [tau, lamb]


def inv_linear_augTikho_sparse(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    nbs=None,
    mu0=None,
    conv_crit=None,
    a0bis=None,
    b0=None,
    a1bis=None,
    b1=None,
    d=None,
    conv_reg=True,
    verb=None,
    verb2head=None,
    maxiter_outer=None,
    maxiter=None,
    tol=None,
    precond=None,       # test
    **kwdargs,
):
    """
    Linear algorithm for Phillips-Tikhonov regularisation
    Called "Augmented Tikhonov", sparese matrix version
    see InvLin_AugTikho_V1.__doc__ for details
    """

    conv = 0.           # convergence variable
    niter = 0           # number of iterations
    mu1 = 0.            # regularisation param

    # verb
    if verb >= 2:
        chi2n = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {chi2n:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*chi2n + mu0*reg:.3e}",
            end='\n',
        )

    # import matplotlib.pyplot as plt # DB
    # import datastock as ds
    # print(R.shape, Tn.shape)
    # print()
    # ds.plot_as_array(R)
    # ds.plot_as_array(Tn)
    # ds.plot_as_array(yn)
    # ds.plot_as_array(sol0)
    # ds.plot_as_array(Tn.dot(sol0))

    # raise Exception()

    # loop
    # Continue until convergence criterion, and at least 2 iterations
    while niter <= 2 or (conv > conv_crit and niter < maxiter_outer):

        # sol = scpsp.linalg.spsolve(
        #    TTn + mu0*R, Tyn,
        #    permc_spec=None,
        #    use_umfpack=True,
        # )

        # seems faster
        sol, itconv = scpsp.linalg.cg(
            TTn + mu0*R, Tyn,
            x0=sol0,
            atol=tol,
            maxiter=maxiter,
            M=precond,
        )

        # call augmented Tikhonov update of mu
        mu1, conv, res2, reg, tau, lamb = _augTikho_update(
            Tn, sol, yn, R,
            a0bis, b0, a1bis, b1,
            d, mu0, conv_reg, nbs, sol0,
            # verb
            verb=verb,
            nchan=nchan,
            niter=niter,
        )

        sol0[:] = sol[:]            # Update reference solution
        niter += 1                  # Update number of iterations
        mu0 = mu1

    return sol, mu1, res2/nchan, reg, niter, [tau, lamb]


def inv_linear_augTikho_chol_dense(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    nbs=None,
    mu0=None,
    conv_crit=None,
    a0bis=None,
    b0=None,
    a1bis=None,
    b1=None,
    d=None,
    conv_reg=True,
    verb=None,
    verb2head=None,
    maxiter_outer=None,
    **kwdargs,
):
    """
    """

    conv = 0.           # convergence variable
    niter = 0           # number of iterations
    mu1 = 0.            # regularisation param

    # verb
    if verb >= 2:
        chi2n = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {chi2n:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*chi2n + mu0*reg:.3e}",
            end='\n',
        )

    # loop
    # Continue until convergence criterion, and at least 2 iterations
    while niter <= 2 or (conv > conv_crit and niter < maxiter_outer):
        try:
            # choleski decomposition requires det(TT + mu0*LL) != 0
            # (chol(A).T * chol(A) = A
            chol = scplin.cholesky(
                TTn + mu0*R,
                lower=False,
                check_finite=False,
                overwrite_a=False,
            )
            # Use np.linalg.lstsq for double-solving the equation
            sol = scplin.cho_solve(
                (chol, False), Tyn,
                overwrite_b=None,
                check_finite=True,
            )
        except Exception as err:
            # call solver
            sol = scplin.solve(
                TTn + mu0*R, Tyn,
                assume_a='sym',         # chol failed => not 'pos'
                overwrite_a=True,       # no significant gain
                overwrite_b=False,      # True faster, but a copy of Tyn needed
                check_finite=False,     # small speed gain compared to True
                transposed=False,
            )  # 3

        # call augmented Tikhonov update of mu
        mu1, conv, res2, reg, tau, lamb = _augTikho_update(
            Tn, sol, yn, R,
            a0bis, b0, a1bis, b1,
            d, mu0, conv_reg, nbs, sol0,
            # verb
            verb=verb,
            nchan=nchan,
            niter=niter,
        )

        # update sol0, mu0 for next iteration
        sol0[:] = sol[:]
        mu0 = mu1
        niter += 1

    return sol, mu1, res2/nchan, reg, niter, [tau, lamb]


def inv_linear_augTikho_chol_sparse(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    nbs=None,
    mu0=None,
    conv_crit=None,
    a0bis=None,
    b0=None,
    a1bis=None,
    b1=None,
    d=None,
    conv_reg=True,
    verb=None,
    verb2head=None,
    maxiter_outer=None,
    **kwdargs,
):

    conv = 0.           # convergence variable
    niter = 0           # number of iterations
    mu1 = 0.            # regularisation param

    # verb
    if verb >= 2:
        chi2n = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {chi2n:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*chi2n + mu0*reg:.3e}",
            end='\n',
        )

    # loop
    # Continue until convergence criterion, and at least 2 iterations
    factor = None
    while niter <= 2 or (conv > conv_crit and niter < maxiter_outer):
        try:
            # choleski decomposition requires det(TT + mu0*LL) != 0
            # A = (chol(A).T * chol(A)
            # optimal if matrix is csc
            if sksp is False:
                factor = scpsp.linalg.factorized(TTn + mu0*R)
                sol = factor(Tyn)
            else:
                if factor is None:
                    factor = sksp.cholmod.cholesky(
                        TTn + mu0*R,
                        beta=0,
                        mode='auto',
                        ordering_method='default',
                        use_long=False,
                    )
                else:
                    # re-use same factor
                    factor.cholesky_inplace(TTn + mu0*R, beta=0)
                sol = factor.solve_A(Tyn)
        except Exception as err:
            # call solver
            sol = scpsp.linalg.spsolve(
                TTn + mu0*R, Tyn,
                permc_spec=None,
                use_umfpack=True,
            )

        # call augmented Tikhonov update of mu
        mu1, conv, res2, reg, tau, lamb = _augTikho_update(
            Tn, sol, yn, R,
            a0bis, b0, a1bis, b1,
            d, mu0, conv_reg, nbs, sol0,
            # verb
            verb=verb,
            nchan=nchan,
            niter=niter,
        )

        # update sol0, mu0 for next iteration
        sol0[:] = sol[:]
        mu0 = mu1
        niter += 1

    return sol, mu1, res2/nchan, reg, niter, [tau, lamb]


def inv_linear_augTikho_pos_dense(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    nbs=None,
    mu0=None,
    conv_crit=None,
    a0bis=None,
    b0=None,
    a1bis=None,
    b1=None,
    d=None,
    conv_reg=True,
    verb=None,
    verb2head=None,
    maxiter_outer=None,
    # specific
    method=None,
    options=None,
    bounds=None,
    func_val=None,
    func_jac=None,
    func_hess=None,
    **kwdargs,
):
    """
    Quadratic algorithm for Phillips-Tikhonov regularisation
    Alternative to the linear version with positivity constraint
    see TFI.InvLin_AugTikho_V1.__doc__ for details
    """

    conv = 0.           # convergence variable
    niter = 0           # number of iterations
    mu1 = 0.            # regularisation param

    # verb
    if verb >= 2:
        chi2n = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {chi2n:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*chi2n + mu0*reg:.3e}",
            end='\n',
        )

    while niter <= 2 or (conv > conv_crit and niter < maxiter_outer):
        # quadratic method for positivity constraint
        sol = scpop.minimize(
            func_val, sol0,
            args=(mu0, Tn, yn, TTn, Tyn, R),
            jac=func_jac,
            hess=func_hess,
            method=method,
            bounds=bounds,
            options=options,
        ).x

        # call augmented Tikhonov update of mu
        mu1, conv, res2, reg, tau, lamb = _augTikho_update(
            Tn, sol, yn, R,
            a0bis, b0, a1bis, b1,
            d, mu0, conv_reg, nbs, sol0,
            # verb
            verb=verb,
            nchan=nchan,
            niter=niter,
        )

        # update sol0, mu0 for next iteration
        sol0[:] = sol[:]
        mu0 = mu1
        niter += 1

    return sol, mu1, res2/nchan, reg, niter, [tau, lamb]


def _augTikho_update(
    Tn, sol, yn, R,
    a0bis, b0, a1bis, b1,
    d, mu0, conv_reg, nbs, sol0,
    # verb
    verb=None,
    nchan=None,
    niter=None,
):
    """
    Linear algorithm for Phillips-Tikhonov regularisation
    Called "Augmented Tikhonov"

    Augmented in the sense that bayesian statistics are combined
        with standard Tikhonov regularisation
    Determines both noise (common multiplicative coefficient) and
        regularisation parameter automatically
    We assume here that all arrays are scaled (noise, conditioning...)
    Sparse matrixes are also prefered to speed-up the computation

    In this method:
      tau is an approximation of the inverse of the noise coefficient
      lamb is an approximation of the regularisation parameter

    N.B.: The noise and reg. param. have probability densities of the form:
        f(x) = x^(a-1) * exp(-bx)
    This function's maximum is in x = (a-1)/b, so a = b+1 gives a maximum at 1.
        (a0, b0) for the reg. param.
        (a1, b1) for the noise estimate

    Ref:
      [1] Jin B., Zou J., Inverse Problems, vol.25, nb.2, 025001, 2009
      [2] http://www.math.uni-bremen.de/zetem/cms/media.php/250/nov14talk_jin%20bangti.pdf
      [3] Kazufumi Ito, Bangti Jin, Jun Zou,
        "A New Choice Rule for Regularization Parameters in Tikhonov
        Regularization", Research report, University of Hong Kong, 2008
    """

    res2 = np.sum(np.power(Tn.dot(sol)-yn, 2))    # residu**2
    reg = sol.dot(R.dot(sol))             # regularity term

    lamb = a0bis/(0.5*reg + b0)           # Update reg. param. estimate
    tau = a1bis/(0.5*res2 + b1)           # Update noise coef. estimate

    # original formula
    mu1 = (lamb/tau) * (2*a1bis/res2)**d  # rescale mu with noise estimate
    # mu1 = (lamb/tau) * (2*a1bis/max(res2, 1e-3))**d  # rescale mu with noise estimate

    if verb >= 3:
        msg = (
            f"\n\t\tconv_reg, d = {conv_reg}, {d}\n"
            f"\t\ta0bis, b0 = {a0bis}, {b0}\n"
            f"\t\ta1bis, b1 = {a1bis}, {b1}\n"
            f"\t\tres2, reg = {res2}, {reg}\n"
            f"\t\tlamb = a0bis/(0.5*reg + b0) = {lamb}\n"
            f"\t\ttau = a1bis/(0.5*res2 + b1) = {tau}\n"
            "\t\tmu = (lamb/tau) * (2*a1bis/res2)**d"
            f" = \t{lamb/tau} * {(2*a1bis/res2)**d} = {mu1}\n"
        )
        print(msg)

    # Compute convergence variable
    if conv_reg:
        conv = np.abs(mu1 - mu0) / mu1
    else:
        sol2 = sol**2
        sol2max = np.max(sol2)
        sol2[sol2 < 0.001*sol2max] = 0.001*sol2max
        conv = np.sqrt(np.sum((sol - sol0)**2 / sol2) / nbs)

    # verb
    if verb >= 2:
        temp1 = f"{nchan} * {res2/nchan:.3e} + {mu1:.3e} * {reg:.3e}"
        temp2 = f"{res2 + mu1*reg:.3e}"
        temp = f"{temp1} = {temp2}"
        print(f"\t\t{niter} \t {temp}   {tau:.3e}   {conv:.3e}")

    return mu1, conv, res2, reg, tau, lamb


# ################################################################
# ################################################################
#               Basic routines - discrepancy principle
# ################################################################


def inv_linear_DisPrinc_sparse(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    mu0=None,
    precond=None,
    verb=None,
    verb2head=None,
    maxiter_outer=None,
    # specific
    chi2n_tol=None,
    chi2n_obj=None,
    maxiter=None,
    tol=None,
    **kwdargs,
):
    """
    Discrepancy principle: find mu such that chi2n = 1 +/- tol
    """

    niter = 0
    lchi2n = np.array([np.sum((Tn.dot(sol0) - yn)**2) / nchan])
    lmu = np.array([mu0])
    chi2n_obj_log = np.log(chi2n_obj)

    # verb
    if verb >= 2:
        reg = sol0.dot(R.dot(sol0))
        temp = f"{nchan} * {lchi2n[0]:.3e} + {mu0:.3e} * {reg:.3e}"
        print(
            f"{verb2head}\n\t\t\t {temp} = {nchan*lchi2n[0] + mu0*reg:.3e}",
            end='\n',
        )

    while niter < 2 or (np.abs(lchi2n[-1] - chi2n_obj) > chi2n_tol and niter < maxiter_outer):

        sol, itconv = scpsp.linalg.cg(
            TTn + lmu[-1]*R, Tyn,
            x0=sol0,
            atol=tol,
            maxiter=maxiter,
            M=precond,
        )

        lchi2n = np.append(lchi2n, np.sum((Tn.dot(sol) - yn)**2) / nchan)

        if niter == 0:
            if lchi2n[-1] >= chi2n_obj + chi2n_tol:
                lmu = np.append(lmu, lmu[-1] / 50.)
            elif lchi2n[-1] <= chi2n_obj - chi2n_tol:
                lmu = np.append(lmu, lmu[-1] * 50.)
            else:
                lmu = np.append(lmu, lmu[-1])
        elif niter == 1 or (
            np.all(lchi2n >= chi2n_obj + chi2n_tol)
            or np.all(lchi2n <= chi2n_obj - chi2n_tol)
        ):
            if lchi2n[-1] >= chi2n_obj + chi2n_tol:
                lmu = np.append(lmu, lmu[-1] / 50.)
            else:
                lmu = np.append(lmu, lmu[-1] * 50.)
        else:
            if lmu[-2] == lmu[-1]:
                # if the algo is stuck => break to avoid infinite loop
                ind = np.argmin(lchi2n[1:] - chi2n_obj)
                lmu[-1] = lmu[ind]
                lchi2n[-1] = lchi2n[ind]
                sol, itconv = scpsp.linalg.cg(
                    TTn + lmu[-1]*R, Tyn,
                    x0=sol0,
                    atol=tol,
                    maxiter=maxiter,
                    M=precond,
                )
                break
            else:
                indsort = np.argsort(lchi2n[1:])
                lmu = np.append(lmu, np.exp(np.interp(
                    chi2n_obj_log,
                    np.log(lchi2n[1:])[indsort],
                    np.log(lmu)[indsort]
                )))

        # verb
        if verb >= 2:
            reg = sol.dot(R.dot(sol))
            res2 = np.sum((Tn.dot(sol)-yn)**2)
            temp1 = f"{nchan} * {lchi2n[-1]:.3e} + {lmu[-1]:.3e} * {reg:.3e}"
            temp2 = f"{res2 + lmu[-1]*reg:.3e}"
            temp = f"{temp1} = {temp2}"
            print(f"\t\t{niter} \t {temp}")
            print(f"\t\t{niter} \t {temp}   \t   {np.abs(lchi2n[-1] - chi2n_obj):.3e}")

        sol0[:] = sol
        niter += 1

    reg = sol.dot(R.dot(sol))           # regularity term
    return sol, lmu[-1], lchi2n[-1], reg, niter, None


# #############################################################################
# #############################################################################
#               Non-regularized - least squares
# #############################################################################


def inv_linear_leastsquares_bounds(
    Tn=None,
    TTn=None,
    Tyn=None,
    R=None,
    yn=None,
    sol0=None,
    nchan=None,
    mu0=None,       # useless
    precond=None,
    verb=None,
    verb2head=None,
    # specific
    A=None,
    b=None,
    maxiter=None,
    tol=None,
    lsmr_tol=None,
    lsq_solver=None,
    method=None,
    bounds=None,
    dconstraints=None,
    **kwdargs,
):
    """
    Discrepancy principle: find mu such that chi2n = 1 +/- tol
    """

    # ---------
    # check inputs

    verbscp = verb
    if method is None:
        method = 'trf'
    if lsq_solver is None:
        if scpsp.issparse(Tn):
            lsq_solver = 'lsmr'
        else:
            lsq_solver = 'exact'

    # ------------
    # optimization

    # verb
    if verb >= 2:
        chi2n0 = np.sum((Tn.dot(sol0) - yn)**2) / nchan
        print(
            f"{verb2head}\n\t\t\t {nchan} * {chi2n0:.3e} = {nchan*chi2n0:.3e}",
            end='\n',
        )

    # optimize
    res = scpop.lsq_linear(
        Tn,
        yn,
        bounds=bounds,
        method=method,
        tol=tol,
        lsq_solver=lsq_solver,
        lsmr_tol=lsmr_tol,
        max_iter=maxiter,
        verbose=verb,
    )

    # --------
    # return

    chi2n = np.sum((Tn.dot(res.x) - yn)**2) / nchan

    # verb
    if verb >= 2:
        temp = f"{nchan} * {chi2n:.3e} = {nchan * chi2n:.3e}"
        print(f"\t\t{res.nit} \t {temp}")

    if dconstraints is None:
        return res.x, None, chi2n, None, res.nit, None
    else:
        return (
            dconstraints['coefs'].dot(res.x) + dconstraints['offset']
        )


# ##################################################################
# ##################################################################
#               _DALGO at import time
# ##################################################################


_DALGO = get_available_inversions_algo(returnas=dict, verb=False)


# -*- coding: utf-8 -*-
"""
Created on Mon Jan 20 11:49:55 2025

@author: dvezinet
"""

import numpy as np


# Useful scalar types
_DTYPES = {
    'nint': (np.int32, np.int64),
    'nfloat': (np.float32, np.float64),
    'bool': (bool, np.bool_),
    'str': (str, np.str_),
}

_DTYPES['int'] = _DTYPES['nint'] + (int,)
_DTYPES['float'] = _DTYPES['nfloat'] + (float,)
_DTYPES['scalar'] = _DTYPES['int'] + _DTYPES['float']